# -*- coding: utf-8 -*-
"""
@file
@brief Various functions about :epkg:`blockdiag`.

.. versionadded:: 1.1
"""
import os
import io
import glob


def _create_fontmap(fontmap, font):
    """
    Inspired from :epkg:`blockdiag` source file (*_bootstrap.py*).
    """
    from blockdiag.utils.fontmap import FontMap
    fontmap = FontMap(fontmap)
    if fontmap.find().path is None or font:
        fontpath = _detectfont(font)
        fontmap.set_default_font(fontpath)
    return fontmap


def _build_diagram(module, builder, drawer, tree, type, code,
                   fontmap=None, antialias=True, nodoctype=False,
                   transparency=False, size=None):
    """
    Inspired from :epkg:`blockdiag` source file (*_bootstrap.py*).
    """
    ScreenNodeBuilder = builder.ScreenNodeBuilder
    diagram = ScreenNodeBuilder.build(tree, None)

    DiagramDraw = drawer.DiagramDraw
    drawer = DiagramDraw(type, diagram, fontmap=fontmap,
                         code=code, antialias=antialias,
                         nodoctype=nodoctype, transparency=transparency)
    drawer.draw()

    if size:
        return drawer.save(size=size)
    else:
        return drawer.save()


def _detectfont(font):
    fontdirs = [
        '/usr/share/fonts',
        '/Library/Fonts',
        '/System/Library/Fonts',
        'c:/windows/fonts',
        '/usr/local/share/font-*',
    ]
    fontfiles = [
        'ipagp.ttf',
        'ipagp.otf',
        'VL-PGothic-Regular.ttf',
        'Hiragino Sans GB W3.otf',
        'AppleGothic.ttf',
        'msgothic.ttf',
        'msgoth04.ttf',
        'msgothic.ttc',
    ]

    fontpath = None
    if font:
        from blockdiag.utils.fontmap import parse_fontpath
        for path in font:
            _path, _ = parse_fontpath(path)
            if os.path.isfile(_path):
                fontpath = path
                break
        else:
            msg = 'fontfile is not found: %s' % font
            raise RuntimeError(msg)

    if fontpath is None:
        globber = (glob.glob(d) for d in fontdirs)
        for fontdir in sum(globber, []):
            for root, _, files in os.walk(fontdir):
                for font_ in fontfiles:
                    if font_ in files:
                        fontpath = os.path.join(root, font_)
                        break

    return fontpath


def draw_diagram(graph, module="blockdiag", format="pillow", **options):
    """
    Draws a graph based on module :epkg:`blockdiag`.

    @param      graph       graph definition,
                            see `syntax <http://blockdiag.com/en/blockdiag/examples.html>`_
    @param      module      ``'blockdiag'`` (only available value)
    @param      format      can be a filename or a module name (``'pillow'``)
    @param      options     additional options for :epkg:`blockdiag`
    @return                 graph

    ::

        blockdiag {
            A -> B -> C -> D;
            A -> E -> F -> G;
        }

    See notebook :ref:`drawdiagramrst`.
    """
    if module == "blockdiag":
        import blockdiag  # pylint: disable=C0415
        module = blockdiag
        import blockdiag.parser  # pylint: disable=C0415
        parser = blockdiag.parser
        import blockdiag.builder  # pylint: disable=C0415
        builder = blockdiag.builder
        import blockdiag.drawer  # pylint: disable=C0415
        drawer = blockdiag.drawer
    else:
        raise ValueError(
            "Unexected value for 'blockdiag': '{0}'".format(module))

    if format in ("pillow", "png"):
        ftype = "png"
    elif format == "svg":
        ftype = "svg"
    else:
        raise ValueError("format should in ['pillow', 'svg']")

    fontmap = _create_fontmap(fontmap=options.get('fontmap', None),
                              font=options.get('font', None))
    tree = parser.parse_string(graph)
    res = _build_diagram(module=module, builder=builder, drawer=drawer,
                         tree=tree, code=graph, fontmap=fontmap, type=ftype,
                         antialias=options.get('antialias', True),
                         nodoctype=options.get('nodoctype', False),
                         transparency=options.get('transparency', False),
                         size=options.get('size', None))
    if format == "pillow":
        from PIL import Image
        image = Image.open(io.BytesIO(res))
        return image
    else:
        return res



"""
@file
@brief CorrPlot functionalities.

It comes from `corrplot.py <https://raw.githubusercontent.com/biokit/biokit/master/biokit/viz/corrplot.py>`_
which I copied here because the module does not properly work on Python 3 (import issues).
See also `biokit license <https://github.com/biokit/biokit/blob/master/LICENSE>`_.

:author: Thomas Cokelaer
:references: http://cran.r-project.org/web/packages/corrplot/vignettes/corrplot-intro.html
"""
import numpy
from scipy.cluster.hierarchy import dendrogram, fcluster
import pandas
from .linkage import Linkage
from ._colormap import cmap_builder


class Corrplot(Linkage):
    """
    An implementation of correlation plotting tools (corrplot).
    the class requires `scipy <http://www.scipy.org/>`_.

    Here is a simple example with a correlation matrix as an input (stored in
    a pandas dataframe):

    .. plot::
        :width: 50%
        :include-source:

        import pandas
        import numpy
        letters = "ABCDEFGHIJKLM"[0:10]
        df = pandas.DataFrame(dict(( (k, numpy.random.random(10)+ord(k)-65) for k in letters)))

        import matplotlib.pyplot as plt
        plt.close('all')
        plt.style.use('ggplot')

        from pyensae.graph_helper import Corrplot
        c = Corrplot(df)
        c.plot()
        plt.show()

    This class requires module `colormap <https://pypi.python.org/pypi/colormap>`_.
    """

    def __init__(self, data, na=0):
        """
        Plots the content of square matrix that contains correlation values.

        :param data: input can be a dataframe (Pandas), or list of lists (python) or
            a numpy matrix. Note, however, that values must be between -1 and 1. If not,
            or if the matrix (or list of lists) is not squared, then correlation is
            computed. The data or computed correlation is stored in :attr:`df` attribute.
        :param na: replace NA values with this value (default 0)

        The :attr:`params` contains some tunable parameters for the colorbar in the
        :meth:`plot` method.

        ::

            # can be a list of lists, the correlation matrix is then a 2x2 matrix
            c = Corrplot([[1,1], [2,4], [3,3], [4,4]])

        """
        super(Corrplot, self).__init__()  # pylint: disable=R1725

        # we delay import in case this is not needed
        self.cmap_builder = cmap_builder

        self.df = pandas.DataFrame(data, copy=True)

        compute_correlation = False

        w, h = self.df.shape
        if self.df.max().max() > 1 or self.df.min().min() < -1:
            compute_correlation = True
        if w != h:
            compute_correlation = True
        if list(self.df.index) != list(self.df.columns):
            compute_correlation = True

        if compute_correlation:
            cor = self.df.corr()
            self.df = cor

        # replace NA with zero
        self.df.fillna(na, inplace=True)

        #: tunable parameters for the :meth:`plot` method.
        self.params = {
            'colorbar.N': 100,
            'colorbar.shrink': .8,
            'colorbar.orientation': 'vertical'}

    def _set_default_cmap(self):
        self.cm = self.cmap_builder('#AA0000', 'white', 'darkblue')

    def order(self, method='complete', metric='euclidean', inplace=False):
        """
        Rearranges the order of rows and columns after clustering.

        :param method: any scipy method (e.g., single, average, centroid,
            median, ward). See scipy.cluster.hierarchy.linkage
        :param metric: any scipy distance (euclidean, hamming, jaccard)
            See scipy.spatial.distance or scipy.cluster.hieararchy
        :param bool inplace: if set to True, the dataframe is replaced

        You probably do not need to use that method. Use :meth:`plot` and
        the two parameters order_metric and order_method instead.
        """
        if None is method or None is metric:
            return self.df
        Y = self.linkage(self.df, method=method, metric=metric)
        ind1 = fcluster(Y, 0.7 * max(Y[:, 2]), 'distance')
        Z = dendrogram(Y, no_plot=True)
        idx1 = Z['leaves']
        cor2 = self.df.iloc[idx1].T.iloc[idx1].T
        if inplace is True:
            self.df = cor2
        else:
            return cor2
        self.Y = Y
        self.Z = Z
        self.idx1 = idx1
        self.ind1 = ind1
        return cor2

    def plot(self, fig=None, grid=True,
             rotation=30, lower=None, upper=None,
             shrink=0.9, axisbg='white', colorbar=True, label_color='black',
             fontsize='small', edgecolor='black', method='ellipse',
             order_method='complete', order_metric='euclidean', cmap=None,
             ax=None, binarise_color=False, figsize=None):
        """
        Plots the correlation matrix from the content of :attr:`df`
        (dataframe).

        By default, the correlation is shown on the upper and lower triangle and is
        symmetric wrt to the diagonal. The symbols are ellipses. The symbols can
        be changed to e.g. rectangle. The symbols are shown on upper and lower sides but
        you could choose a symbol for the upper side and another for the lower side using
        the **lower** and **upper** parameters.

        :param fig: Create a new figure by default. If an instance of an existing
            figure is provided, the corrplot is overlayed on the figure provided.
            Can also be the number of the figure.
        :param grid: add grid (Defaults to grey color). You can set it to False or a color.
        :param rotation: rotate labels on y-axis
        :param lower: if set to a valid method, plots the data on the lower
            left triangle
        :param upper: if set to a valid method, plots the data on the upper
            left triangle
        :param float shrink: maximum space used (in percent) by a symbol.
            If negative values are provided, the absolute value is taken.
            If greater than 1, the symbols wiill overlap.
        :param axisbg: color of the background (defaults to white).
        :param colorbar: add the colorbar (defaults to True).
        :param str label_color: (defaults to black).
        :param fontsize: size of the fonts defaults to 'small'.
        :param method: shape to be used in 'ellipse', 'square', 'rectangle',
            'color', 'text', 'circle',  'number', 'pie'.

        :param order_method: see :meth:`order`.
        :param order_metric: see : meth:`order`.
        :param cmap: a valid cmap from matplotlib or colormap package (e.g.,
            'jet', or 'copper'). Default is red/white/blue colors.
        :param binarise_color: two colors only, negative, positive
        :param ax: a matplotlib axes.
        :param figsize: gives that parameter to the new created figure
        :return: ax (matplotlib axes)

        The colorbar can be tuned with the parameters stored in :attr:`params`.
        Here is an example. See notebook for other examples:

        ::

            c = corrplot.Corrplot(dataframe)
            c.plot(cmap=('Orange', 'white', 'green'))
            c.plot(method='circle')
            c.plot(colorbar=False, shrink=.8, upper='circle'  )
        """
        import matplotlib.pyplot as plt  # pylint: disable=C0415

        # default
        if cmap is not None:
            try:
                if isinstance(cmap, str):
                    self.cm = self.cmap_builder(cmap)
                else:
                    self.cm = self.cmap_builder(*cmap)
            except Exception:
                self._set_default_cmap()
        else:
            self._set_default_cmap()

        self.shrink = abs(shrink)
        self.fontsize = fontsize
        self.edgecolor = edgecolor

        df = self.order(method=order_method, metric=order_metric)

        # figure can be a number or an instance; otherwise creates it
        params = dict(facecolor=axisbg)
        if isinstance(fig, int):
            params["num"] = fig.number
        elif fig is not None:
            params["num"] = fig.number
        else:
            params["num"] = None
        if figsize is not None:
            params["figsize"] = figsize
        fig = plt.figure(**params)

        # do we have an axes to plot the data in ?
        if ax is None:
            ax = plt.subplot(1, 1, 1, aspect='equal', facecolor=axisbg)
        else:
            # if so, clear the axes. Colorbar cannot be removed easily.
            plt.sca(ax)
            ax.clear()

        # subplot resets the bg color, let us set it again
        fig.set_facecolor(axisbg)

        width, height = df.shape
        labels = (df.columns)

        if upper is None and lower is None:
            mode = 'method'
        elif upper and lower:
            mode = 'both'
        elif lower is not None:
            mode = 'lower'
        elif upper is not None:
            mode = 'upper'

        self.binarise_color = binarise_color
        if mode == 'upper':
            self._add_patches(df, upper, 'upper', ax, diagonal=True)
        elif mode == 'lower':
            self._add_patches(df, lower, 'lower', ax, diagonal=True)
        elif mode == 'method':
            self._add_patches(df, method, 'both', ax, diagonal=True)
        elif mode == 'both':
            self._add_patches(df, upper, 'upper', ax, diagonal=False)
            self._add_patches(df, lower, 'lower', ax, diagonal=False)

        # shift the limits to englobe the patches correctly
        ax.set_xlim(-0.5, width - .5)
        ax.set_ylim(-0.5, height - .5)

        # set xticks/xlabels on top
        ax.xaxis.tick_top()
        xtickslocs = numpy.arange(len(labels))
        ax.set_xticks(xtickslocs)
        ax.set_xticklabels(labels, rotation=rotation, color=label_color,
                           fontsize=fontsize, ha='left')

        ax.invert_yaxis()
        ytickslocs = numpy.arange(len(labels))
        ax.set_yticks(ytickslocs)
        ax.set_yticklabels(labels, fontsize=fontsize, color=label_color)
        plt.tight_layout()

        if grid is not False:
            if grid is True:
                grid = 'grey'
            for i in range(0, width):
                ratio1 = float(i) / width
                ratio2 = float(i + 2) / width
                # set axis off
                # 2 - set xlabels along the diagonal
                # set colorbar either on left or bottom
                if mode == 'lower':
                    ax.axvline(i + .5, ymin=1 - ratio1, ymax=0., color=grid)
                    ax.axhline(i + .5, xmin=0, xmax=ratio2, color=grid)
                if mode == 'upper':
                    ax.axvline(i + .5, ymin=1 - ratio2, ymax=1, color=grid)
                    ax.axhline(i + .5, xmin=ratio1, xmax=1, color=grid)
                if mode in ['method', 'both']:
                    ax.axvline(i + .5, color=grid)
                    ax.axhline(i + .5, color=grid)

            # can probably be simplified
            if mode == 'lower':
                ax.axvline(-.5, ymin=0, ymax=1, color='grey')
                ax.axvline(width - .5, ymin=0, ymax=1. /
                           width, color='grey', lw=2)
                ax.axhline(width - .5, xmin=0, xmax=1, color='grey', lw=2)
                ax.axhline(-.5, xmin=0, xmax=1. / width, color='grey', lw=2)
                ax.xticks([])
                for i in range(0, width):
                    ax.text(i, i - .6, labels[i], fontsize=fontsize,
                            color=label_color,
                            rotation=rotation, verticalalignment='bottom')
                    ax.text(-.6, i, labels[i], fontsize=fontsize,
                            color=label_color,
                            rotation=0, horizontalalignment='right')
                ax.set_axis_off()
            # can probably be simplified
            elif mode == 'upper':
                ax.axvline(width - .5, ymin=0, ymax=1, color='grey', lw=2)
                ax.axvline(-.5, ymin=1 - 1. / width,
                           ymax=1, color='grey', lw=2)
                ax.axhline(-.5, xmin=0, xmax=1, color='grey', lw=2)
                ax.axhline(width - .5, xmin=1 - 1. / width,
                           xmax=1, color='grey', lw=2)
                ax.yticks([])
                for i in range(0, width):
                    ax.text(-.6 + i, i, labels[i], fontsize=fontsize,
                            color=label_color, horizontalalignment='right',
                            rotation=0)
                    ax.text(i, -.5, labels[i], fontsize=fontsize,
                            color=label_color, rotation=rotation, verticalalignment='bottom')
                ax.set_axis_off()

        # set all ticks length to zero
        ax = plt.gca()
        ax.tick_params(axis='both', which='both', length=0)

        if colorbar:
            N = self.params['colorbar.N'] + 1
            if N < 2:
                raise RuntimeError("No colorbar to draw.")
            cb = plt.gcf().colorbar(
                self.collection, orientation=self.params['colorbar.orientation'],
                shrink=self.params['colorbar.shrink'],
                boundaries=numpy.linspace(0, 1, N),
                ticks=[0, .25, 0.5, 0.75, 1])
            cb.ax.set_yticklabels([-1, -.5, 0, .5, 1])
            # make sure it goes from -1 to 1 even though actual values may not
            # reach that range
            # cb.set_clim(0, 1)
            # not working in matplotlib 3.3.0

        return ax

    def _add_patches(self, df, method, fill, ax, diagonal=True):

        from matplotlib.patches import Ellipse, Circle, Rectangle, Wedge
        from matplotlib.collections import PatchCollection

        width, height = df.shape

        patches = []
        colors = []
        for x in range(width):
            for y in range(height):
                if fill == 'lower' and x > y:
                    continue
                if fill == 'upper' and x < y:
                    continue
                if diagonal is False and x == y:
                    continue
                datum = (df.iloc[x, y] + 1.) / 2.
                d = df.iloc[x, y]
                d_abs = numpy.abs(d)
                #c = self.pvalues[x, y]
                rotate = -45 if d > 0 else +45
                #cmap = self.poscm if d >= 0 else self.negcm
                if method in ['ellipse', 'square', 'rectangle', 'color']:
                    if method == 'ellipse':
                        func = Ellipse
                        patch = func((x, y), width=1 * self.shrink,
                                     height=(self.shrink - d_abs * self.shrink), angle=rotate)
                    else:
                        func = Rectangle
                        w = h = d_abs * self.shrink
                        offset = (1 - w) / 2.
                        if method == 'color':
                            w = 1
                            h = 1
                            offset = 0
                        patch = func((x + offset - .5, y + offset - .5), width=w,
                                     height=h, angle=0)
                    if self.edgecolor:
                        patch.set_edgecolor(self.edgecolor)
                    # patch.set_facecolor(cmap(d_abs))
                    colors.append(datum)
                    if d_abs > 0.05:
                        patch.set_linestyle('dotted')
                    # ax.add_artist(patch)
                    patches.append(patch)
                elif method == 'circle':
                    patch = Circle((x, y), radius=d_abs * self.shrink / 2.)
                    if self.edgecolor:
                        patch.set_edgecolor(self.edgecolor)
                    # patch.set_facecolor(cmap(d_abs))
                    colors.append(datum)
                    if d_abs > 0.05:
                        patch.set_linestyle('dotted')
                    # ax.add_artist(patch)
                    patches.append(patch)
                elif method in ['number', 'text']:
                    if d < 0:
                        edgecolor = self.cm(-1.0)
                    elif d >= 0:
                        edgecolor = self.cm(1.0)
                    d_str = "{:.2f}".format(d).replace(
                        "0.", ".").replace(".00", "")
                    ax.text(x, y, d_str, color=edgecolor,
                            fontsize=self.fontsize, horizontalalignment='center',
                            weight='bold', alpha=max(0.5, d_abs),
                            withdash=False)
                elif method == 'pie':
                    S = 360 * d_abs
                    patch = [
                        Wedge((x, y), 1 * self.shrink / 2., -90, S - 90),
                        Wedge((x, y), 1 * self.shrink / 2., S - 90, 360 - 90),
                    ]
                    # patch[0].set_facecolor(cmap(d_abs))
                    # patch[1].set_facecolor('white')
                    colors.append(datum)
                    colors.append(0.5)
                    if self.edgecolor:
                        patch[0].set_edgecolor(self.edgecolor)
                        patch[1].set_edgecolor(self.edgecolor)

                    # ax.add_artist(patch[0])
                    # ax.add_artist(patch[1])
                    patches.append(patch[0])
                    patches.append(patch[1])
                else:
                    raise ValueError(
                        'Method for the symbols is not known. Use e.g, square, circle')

        if self.binarise_color:
            colors = [1 if color > 0.5 else -1 for color in colors]

        if len(patches):
            col1 = PatchCollection(
                patches, array=numpy.array(colors), cmap=self.cm)
            ax.add_collection(col1)

            self.collection = col1



"""
@file
@brief      generic class to access a SQL database
"""
from pyquickhelper.loghelper import fLOG
from .database_core import DatabaseCore
from .database_import_export import DatabaseImportExport
from .database_object import DatabaseObject
from .database_join_group import DatabaseJoinGroup


class Database(DatabaseCore, DatabaseImportExport, DatabaseObject, DatabaseJoinGroup):
    """
    This class allows the user to load table from text files and store them into a
    SQL file which can be empty or not,
    it is using :epkg:`SQLite3` module.
    Under Windows, you can use
    `SQLiteSpy <http://www.yunqa.de/delphi/doku.php/products/sqlitespy/index>`_
    to have a graphical overview of the database.
    Parameter *dbfile* can be of type
    `sqlite3.Connection <https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection>`_.
    """

    def __init__(self, dbfile, engine="SQLite", user=None, password=None,
                 host="localhost", LOG=fLOG, attach=None):
        """

        @param      dbfile          database file (use ``:memory:`` to avoid creating a file and using only memory)
                                    it can also contain several files separated by ;
                                    ``name_file ; nickname,second_file ; ...``
        @param      engine          SQLite or MySQL (if it is installed)
        @param      user            user if needed
        @param      password        password if needed
        @param      host            to connect to a MSSQL database
        @param      LOG             LOG function
        @param      attach          dictionary: { nickname: filename }, list of database to attach

        @warning If the folder does not exist, it will be created
        """
        DatabaseJoinGroup.__init__(self)
        DatabaseCore.__init__(self, sql_file=dbfile, engine=engine, user=user, password=password,
                              host=host, LOG=LOG, attach=attach)

    @staticmethod
    def schema_database(df, add_id=True):
        """
        Returns the schema for a database which would contains this database.

        @param      df          pandas DataFrame
        @param      add_id      if True, adds an index "PRIMARYKEY"
        @return                 dictionary { index_column: (name, type) }
        """
        schema = {i: (l, str) for i, l in enumerate(df.columns)}
        if add_id is not None:
            if isinstance(add_id, bool):
                if add_id:
                    add_id = "PRIMARYKEY"
                    schema[-1] = (add_id, int, "PRIMARYKEY", "AUTOINCREMENT")
            else:
                schema[-1] = (add_id, int, "PRIMARYKEY", "AUTOINCREMENT")

        if len(df) > 0:
            # we use the first row to determine type
            for i, v in enumerate(df.values[0]):
                if not isinstance(v, str):
                    schema[i] = (schema[i][0], type(v))
        return schema

    @staticmethod
    def fill_sql_table(df, filename_or_database, tablename, add_id="idr", **kwargs):
        """
        Returns a Database object, creates the database if it does not exists,
        same for the table.

        @param      df                      pandas DataFrame
        @param      filename_or_database    filename or Database object,
                                                in that second case, we assume method connect was called before
        @param      tablename               table name
        @param      add_id                  if != None then the function adds an id, it first takes the
                                            ``max(id)`` and goes on incrementing it
        @param      kwargs                  sent to @see cl Database
        @return                             ``Database`` object (new or the one from the parameters),
                                            in both case, the database is not disconnected

        .. exref::
            :title: import a DataFrame into a SQL table
            :tag: SQL

            ::

                values = [  {"name":"A", "age":10, "score":34.5 },
                            {"name":"B", "age":20, "score":-34.5 }, ]
                df  = pandas.DataFrame(values)
                dbf = "something.db3"
                db  = Database.fill_sql_table(df, dbf, "mytable")

            This example could be replaced by:

            ::

                values = [  {"name":"A", "age":10, "score":34.5 },
                            {"name":"B", "age":20, "score":-34.5 }, ]
                df  = pandas.DataFrame(values)
                dbf = "something.db3"
                db  = Database(dbf)
                db.connect()
                db.import_dataframe(df, "mytable)
                db.close()
        """

        schema = Database.schema_database(df, add_id)

        if isinstance(filename_or_database, str):
            db = Database(filename_or_database, **kwargs)
            db.connect()

            if tablename not in db.get_table_list():
                cursor = db.create_table(tablename, schema)
                db.append_values(df.values, tablename, schema, cursor=cursor)
            else:
                db.append_values(df.values, tablename, schema)
        else:
            db = filename_or_database
            if tablename not in db.get_table_list():
                cursor = db.create_table(tablename, schema)
                db.append_values(df.values, tablename, schema, cursor=cursor)
            else:
                db.append_values(df.values, tablename, schema)

        return db

    def import_dataframe(self, df, tablename, add_id="idr"):
        """
        Imports a DataFrame into a table.

        @param      df              pandas DataFrame
        @param      tablename       table name
        @param      add_id          an index, maybe to be added
        @return                     self
        """
        return Database.fill_sql_table(df, self, tablename, add_id)

    def to_df(self, request):
        """
        Converts a SQL request into a :epkg:`pandas:Dataframe`.

        @param      request     SQL request
        @return                 DataFrame
        """
        import pandas  # pylint: disable=C0415
        cols = self.get_sql_columns(request)
        iter = self.execute_view(request, nolog=True)
        return pandas.DataFrame(iter, columns=cols)

    def copy_to(self, db, subset=None):
        """
        Copies all tables into db, we assume both database are not connected.

        @param      db      another database (possibly empty)
        @param      subset  list of tables to copy or None for all
        """
        self.connect()
        db.connect()
        for tbl in self.get_table_list():
            if subset is None or tbl in subset:
                self.LOG("copy_to: create table " + tbl)
                sch = self.get_table_columns_list(tbl, True)
                curins = db.create_table(tbl, sch)
                cursor = self.execute("SELECT * FROM %s" % tbl)
                buffer = []
                for row in cursor:
                    buffer.append(row)
                    if len(buffer) >= 1000:
                        db.insert(tbl, buffer, cursor=curins)
                        buffer = []
                if len(buffer) > 0:
                    db.insert(tbl, buffer)
                db.commit()
                cursor.close()
        self.close()
        db.close()



# coding: latin-1
"""
@file

@brief contains a class which iterations on rows of a text file structured as a table.

"""


import re
import os
import decimal

from pyquickhelper.loghelper import noLOG
from pyquickhelper.loghelper.flog import GetPath
from .file_text_binary import TextFile


class TextFileColumns(TextFile):
    """
    This class opens a text file as if it were a binary file. It can deal with null characters.
    The file is interpreted as a TSV file or file containing columns.
    The separator is found automatically.
    The columns are assumed to be in the first line but it is not mandatory.
    It walks along a file through an iterator, every line is automatically converted into a dictionary ``{ column : value }``.
    If the class was able to guess what type is which column, the conversion will automatically take place.

    ::

        f = TextFileColumns(filename)
                # filename is a file
                # the separator is unknown --> the class automatically determines it
                # as well as the columns and their type
        f.open()
        for d in f:
            print(d)       # d is a dictionary
        f.close()

    @var _force_header      there is a header even if not detected
    @var _force_noheader    there is no header even if detected
    @var _changes           replace the columns name
    @var _regexfix          impose a regular expression to interpret a line instead of the automatically built one
    @var _filter_dict       it is a function which takes a dictionary and returns a boolean which tells if the line must considered or not
    @var _fields            name of the columns (if there is no header)

    Spaces and non-ascii characters cannot be used to name a column.
    This name must be a named group for a regular expression.
    """

    def __init__(self, filename, errors=None, fLOG=noLOG, force_header=False, changes=None,
                 force_noheader=False, regex=None, filter=None, fields=None,
                 keep_text_when_bad_type=False, break_at=-1, strip_space=True,
                 force_sep=None, nb_line_guess=100, mistake=3, encoding="utf-8",
                 strict_separator=False):
        """
        @param      filename                    filename
        @param      errors                      see str (errors = ...)
        @param      fLOG                        LOG function, see `fLOG
                                                <http://www.xavierdupre.fr/app/pyquickhelper/helpsphinx/
                                                pyquickhelper/loghelper/flog.html#pyquickhelper.loghelper.flog.fLOG>`_
        @param      force_header                defines the first line as columns header whatever is it relevant or not
        @param      changes                     to change the column name, gives the correspondence, example: { "query":"query___" },
                                                it can be a list if there is no header and you want to name any column
        @param      force_noheader              there is no header at all
        @param      regex                       specify a different regular expression (only if changes is a list)
                                                if it is a dictionary, the class will replace the default
                                                by the one associated in regex for this field
        @param      filter                      None if there is no filter, otherwise it is a function
                                                which takes a dictionary and returns a boolean
                                                which tells if the line must considered or not
        @param      fields                      when the header is not here, these fields will name the columns
        @param      keep_text_when_bad_type     keep the value when the conversion type does not word
        @param      break_at                    if != -1, stop when this limit is reached
        @param      strip_space                 remove space around columns if True
        @param      force_sep                   if != None, impose a column separator
        @param      nb_line_guess               number of lines used to guess types
        @param      mistake                     not more than mistake conversion in numbers are allowed
        @param      encoding                    encoding
        @param      strict_separator            strict number of columns, it assumes there is no separator in the content of every column
        """
        if changes is None:
            changes = {}

        TextFile.__init__(self, filename, errors, fLOG=fLOG, encoding=encoding)

        self._force_header = force_header
        self._force_noheader = force_noheader
        self._changes = changes
        self._regexfix = regex
        self._filter_dict = filter
        self._fields = fields
        self._keep_text_when_bad_type = keep_text_when_bad_type
        self._break_at = break_at
        self._strip_space = strip_space
        self._force_sep = force_sep
        self._nb_guess_line = nb_line_guess
        self._mistake = mistake
        self._strict_separator = strict_separator
        self._encoding = encoding

        if isinstance(changes, list):
            hhhh, _ = 0, len(changes)
            while _ > 0:
                hhhh, _ = hhhh, _ / 10  # pylint: disable=W0127
            forma_ = "c%0" + str(hhhh) + "d"

            self._changes = {}
            for i, c in enumerate(changes):
                self._changes[forma_ % i] = c

            if self._regexfix is not None and \
                    not isinstance(self._regexfix, dict) and \
                    "(?P<" not in self._regexfix:
                reg = re.compile("[(](.+?)[)]")
                fi = reg.findall(self._regexfix)
                if len(fi) != len(changes):
                    raise Exception(
                        "not the same number of fields in regular expression (%d,%d):\n%s\n%s" %
                        (len(fi), len(changes), str(fi), str(changes)))
                exp = []
                for a, b in zip(fi, changes):
                    s = "(?P<%s>%s)" % (b, a)
                    exp.append(s)
                p = self._regexfix.find(")") + 1
                s = self._regexfix[p]
                self._regexfix = s.join(exp)
                self.LOG("split: ", fi)
                self.LOG("new regex: ", self._regexfix)
            else:
                self.LOG("    TextFileColumns (1): regex: ", self._regexfix)
        else:
            self.LOG("    TextFileColumns (2): regex: ", self._regexfix)

    def __str__(self):
        """
        Returns the header.
        """
        return str(self.__dict__)

    def get_columns(self):
        """
        @return         the columns
        """
        if "_columns" not in self.__dict__:
            raise Exception("there is no available columns")
        return self._columns

    def open(self):
        """
        Opens the file and find out if there is a header, what are the columns, what are their type...
        any information about which format was found is logged.
        """
        if "_header" not in self.__dict__:
            header, columns, sep, regex = self.guess_columns(force_header=self._force_header,
                                                             changes=self._changes,
                                                             force_noheader=self._force_noheader,
                                                             fields=self._fields,
                                                             regex=self._regexfix if isinstance(
                                                                 self._regexfix,
                                                                 dict) else {},
                                                             force_sep=self._force_sep,
                                                             nb=self._nb_guess_line,
                                                             mistake=self._mistake)
            if self._regexfix is not None and not isinstance(
                    self._regexfix, dict):
                regex = self._regexfix
            self._header = header
            self._columns = columns
            self._sep = sep
            try:
                self._regex = re.compile(regex)
            except Exception as e:
                raise RuntimeError(  # pylint: disable=W0707
                    "algorithm problem: (type %r, %r)\nunable to understand a regular expression (file %r)\nexp: %r" %
                    (str(type(e)), str(e), self.filename, regex))
            self._name = {}
            self._nb = 0
            self._conv = {}
            for k, v in self._columns.items():
                self._name[v[0]] = (k, v[1])
                if v[1] in [int, float, decimal.Decimal]:
                    self._conv[v[0]] = v[1]
        self._nb += 1
        TextFile.open(self)

    def close(self):
        """
        Closes the file and remove all information related to the format,
        next time it is opened, the format will be checked again.
        """
        TextFile.close(self)
        self._nb -= 1
        if self._nb == 0:
            del self.__dict__["_header"]
            del self.__dict__["_columns"]
            del self.__dict__["_regex"]
            del self.__dict__["_name"]
            del self.__dict__["_conv"]

    def __iter__(self):
        """
        @return         a dictionary ``{ column_name: value }``
        """
        class tempo__:

            def __init__(self, r):
                self.res = r

            def groupdict(self):
                return self.res

        if "_header" not in self.__dict__:
            raise Exception("file not open %s" % self.filename)

        regex_simple = re.compile(self._regex.pattern.replace(">.*)", ">.*?)"))

        nb = 0
        nberr = 0
        nbert = 0
        for line in TextFile.__iter__(self):
            if nb == 0 and self._header:
                nb += 1
                continue

            tempc = line.split(self._sep)

            if len(tempc) == len(self._columns):
                res = {}
                for i, a in enumerate(tempc):
                    res[self._columns[i][0]] = a
                r = tempo__(res)
            elif not self._strict_separator:
                if len(tempc) < len(self._columns):
                    # impossible
                    r = None
                else:
                    # conflicts...
                    r = regex_simple.match(line)
                    if r is None:
                        r = self._regex.match(line)
            else:
                r = None

            if r is None:
                if nberr == 0:
                    self.LOG(self._regex.pattern)
                self.LOG(
                    "error regex",
                    nberr,
                    "unable to interpret line ",
                    nb,
                    ": ",
                    repr(line))
                nberr += 1
                if nberr * 10 > nb and nberr > 4:
                    message = "pattern: %s\n line: %s" % (
                        regex_simple.pattern, line)
                    raise Exception(
                        "(a) there are probably too many errors %d (%d)\n%s" %
                        (nberr, nb, message))
            else:
                res = r.groupdict()
                if self._strip_space:
                    for k in res:
                        res[k] = res[k].strip()
                giveup = False

                for k in res:
                    if k in self._conv:
                        try:
                            if len(res[k]) == 0 and (self._conv[k] == int or self._conv[
                                    k] == float or self._conv[k] == decimal.Decimal):
                                ttt = self._conv[k](0)
                            else:
                                ttt = self._conv[k](res[k])
                            res[k] = ttt
                        except ValueError:
                            nbert += 1
                            if self._keep_text_when_bad_type:
                                if nbert % 1000 == 1:
                                    self.LOG(
                                        "error type",
                                        nbert,
                                        "unable to interpret line (but keep it) ",
                                        nb,
                                        "value",
                                        repr(
                                            res[k]),
                                        " type ",
                                        repr(
                                            self._conv[k]),
                                        " line ",
                                        repr(line))
                            else:
                                self.LOG(
                                    "error type", nbert, "unable to interpret line ", nb, "value", repr(
                                        res[k]), " type ", repr(
                                        self._conv[k]), " line ", repr(line))
                                if nbert * 10 > nb and nbert > 4:
                                    message = "pattern: %s\n line: %s" % (
                                        regex_simple.pattern, line)
                                    raise RuntimeError(  # pylint: disable=W0707
                                        "(b) there are probably too many errors %r\n%r" %
                                        (nberr, message))
                                giveup = True
                                break
                if giveup:
                    continue
                if self._filter_dict is None or self._filter_dict(res):
                    yield res

            nb += 1
            if self._break_at != -1 and nb > self._break_at:
                break

    @staticmethod
    def _store(output, la, encoding="utf-8"):
        """
        Stores a list of dictionaries into a file (add a header).

        @param      output      filename
        @param      la          list of dictionary key:value
        @param      encoding    encoding
        @warning                format is utf-8
        """
        sepline = "\n"  # GetSepLine ()
        f = open(output, "w", encoding=encoding)
        nbline = 0
        for d in la:
            if nbline == 0:
                keys = list(d.keys())
                keys.sort()
                f.write("\t".join(keys) + sepline)

            val = [str(d[k]) for k in keys]
            s = "\t".join(val)
            f.write(s + sepline)

            nbline += 1
        f.close()

    def sort(self, output, key, maxmemory=2 ** 28, folder=None, fLOG=noLOG):
        """
        Sorts a text file, even a big one, one or several columns gives the order.

        @param      output      output file result
        @param      key         lines sorted depending of these columns
        @param      maxmemory   a file is split into smaller files which contains not more than maxmemory lines
        @param      folder      the function needs to create temporary files, this folder will contain them
                                before they get removed
        @param      fLOG        logging function
        @return

        @warning  We assume this file is not opened.
        """
        if isinstance(key, str):
            key = (key,)
        if folder is None:
            folder = GetPath()
        if not os.path.exists(folder):
            raise Exception("unable to find folder %s" % folder)

        try:
            file = open(output, "w", encoding=self._encoding)
            file.close()
        except Exception as e:
            raise RuntimeError(  # pylint: disable=W0707
                "Unable to create file %r, reason: %r" %
                (output, str(e)))

        self.LOG("sorting file ", self.filename)
        #root   = self.filename.replace (":", "_").replace ("/", "_").replace ("\\", "_").replace (".", "_")
        files = []
        memo = []
        self.open()
        for line in self:
            try:
                k = tuple(line[k] for k in key)
            except KeyError as e:
                raise Exception("unable to find one column in\n{0}".format(
                    self.get_columns())) from e
            memo.append((k, line))
            if len(memo) > maxmemory:
                memo.sort(key=lambda el: el[0])
                memo = [la[1] for la in memo]
                tempout = os.path.join(folder, "root_%05d.txt" % len(files))
                self.LOG("writing file %d lines in " % len(memo), tempout)
                TextFileColumns._store(tempout, memo)
                files.append(tempout)
                memo = []

        if len(memo) > 0:
            memo.sort(key=lambda el: el[0])
            memo = [la[1] for la in memo]
            tempout = os.path.join(folder, "root_%05d.txt" % len(files))
            self.LOG("writing file %d lines in " % len(memo), tempout)
            TextFileColumns._store(tempout, memo)
            files.append(tempout)
            memo = []

        self.close()

        TextFileColumns.fusion(
            key,
            files,
            output,
            force_header=self._force_header,
            fLOG=self.LOG)
        for m in files:
            self.LOG("removing ", m)
            os.remove(m)

    @staticmethod
    def fusion(key, files, output, force_header=False, encoding="utf-8", fLOG=noLOG):
        """
        Does a fusion between several files
        with the same columns (different order is allowed).

        @param      key             columns to be compared
        @param      files           list of files
        @param      output          output file
        @param      force_header    impose the first line as a header
        @param      encoding        encoding
        @param      fLOG            logging function
        @warning We assume all files are sorted depending on columns in key
        """
        fh = []
        for f in files:
            h = TextFileColumns(f, force_header=force_header,
                                encoding=encoding, fLOG=fLOG)
            h.open()
            fh.append([h, iter(h)])

        res = open(output, "w", encoding=encoding)
        nbline = 0
        sepline = "\n"  # GetSepLine ()
        if isinstance(key, str):
            key = [key]

        # start
        kline = []
        for li in fh:
            try:
                if li[1] is None:
                    d = None
                else:
                    d = li[1].__next__()
            except StopIteration:
                d = None
            if d is not None:
                try:
                    k = tuple(d[k] for k in key)
                except KeyError as e:
                    raise Exception("unable to find one column in\n{0}".format(
                        li[0].get_columns())) from e
                kline.append([k, d] + li)

        # loop
        while len(kline) > 0:

            # minimum
            mi = None
            for i, line in enumerate(kline):
                if mi is None or line[0] < mi:
                    mi = line[0]
                    pos = i

            # picking
            line = kline[pos]
            del kline[pos]

            # adding
            d = line[1]
            if nbline == 0:
                keys = list(d.keys())
                keys.sort()
                res.write("\t".join(keys) + sepline)

            val = [str(d[k_]) for k_ in keys]
            s = "\t".join(val)
            res.write(s + sepline)
            nbline += 1

            # next
            try:
                d = line[-1].__next__()
            except StopIteration:
                d = None

            if d is not None:
                k = tuple(d[k_] for k_ in key)
                kline.append([k, d] + line[2:])

        # end
        for li in fh:
            li[0].close()
        res.close()



# -*- coding: utf-8 -*-
"""
@file
@brief Defines SQL commands to play with `sqlite3 <https://docs.python.org/3/library/sqlite3.html>`_.
See notebook :ref:`pyensaesqlmagicrst`.
"""
import os

from IPython.core.magic import magics_class, line_magic
from IPython.core.magic import line_cell_magic
from pyquickhelper.ipythonhelper import MagicCommandParser, MagicClassWithHelpers
from .sql_interface import InterfaceSQL, InterfaceSQLException


@magics_class
class MagicSQL(MagicClassWithHelpers):

    """
    Defines SQL commands to play with `sqlite3 <https://docs.python.org/3.4/library/sqlite3.html>`_
    See notebook :ref:`pyensaesqlmagicrst`.
    """

    def get_connection(self, name):
        """
        returns the connection stored in the workspace

        @param      name        variable name of the database
        @return                 object
        """
        if isinstance(name, str):
            if self.shell is None:
                raise Exception("No detected workspace.")

            if name not in self.shell.user_ns:
                raise KeyError(
                    "No opened sqlite3 database called: " + str(name))

            res = self.shell.user_ns[name]
        else:
            res = name
        if not isinstance(res, InterfaceSQL):
            res = InterfaceSQL.create(res)
        return res

    @staticmethod
    def SQL_connect_parser():
        """
        defines the way to parse the magic command ``%SQL_connect``
        """
        parser = MagicCommandParser(prog="SQL_connect",
                                    description='connect to a SQL database')
        parser.add_argument('filename', type=str,
                            help='database filename', eval_type=str)
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object')
        return parser

    @line_magic
    def SQL_connect(self, line):
        """
        define ``SQL_connect`` which connects to a SQL database,
        it stores the database object in variable DB by default

        .. nbref::
            :tag: SQL
            :title: SQL_connect_parser

            The code for magic command ``%SQL_connect_parser`` is equivalent to::

                from pyense.sql import InterfaceSQL
                obj = InterfaceSQL.create(args.filename)
                obj.connect()

            See notebook :ref:`pyensaesqlmagicrst`.
        """
        parser = self.get_parser(MagicSQL.SQL_connect_parser, "SQL_connect")
        args = self.get_args(line, parser)

        if args is not None:
            obj = InterfaceSQL.create(args.filename)
            obj.connect()
            self.shell.user_ns[args.variable] = obj
            return obj

    @staticmethod
    def SQL_close_parser():
        """
        defines the way to parse the magic command ``%SQL_close``
        """
        parser = MagicCommandParser(prog="SQL_close",
                                    description='connect to a SQL database')
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object (and to close)')
        return parser

    @line_magic
    def SQL_close(self, line=""):
        """
        define ``SQL_close`` which closes a database

        .. nbref::
            :tag: SQL
            :title: SQL_close

            The code for magic command ``%SQL_close`` is equivalent to::

                db.close()

            See notebook :ref:`pyensaesqlmagicrst`.
        """
        parser = self.get_parser(MagicSQL.SQL_close_parser, "SQL_close")
        args = self.get_args(line, parser)

        if args is not None:
            db = self.get_connection(args.variable)
            r = db.close()
            return r

    @staticmethod
    def SQL_tables_parser():
        """
        defines the way to parse the magic command ``%SQL_tables``
        """
        parser = MagicCommandParser(prog="SQL_tables",
                                    description='list the tables of a database')
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object')
        return parser

    @line_magic
    def SQL_tables(self, line=""):
        """
        define ``%SQL_tables`` whichs lists the tables in a database

        .. nbref::
            :tag: SQL
            :title: SQL_tables

            The code for magic command ``%SQL_tables`` is equivalent to::

                db.get_table_list()

            See notebook :ref:`pyensaesqlmagicrst`.
        """
        parser = self.get_parser(MagicSQL.SQL_tables_parser, "SQL_tables")
        args = self.get_args(line, parser)

        if args is not None:
            db = self.get_connection(args.variable)
            return db.get_table_list()

    @staticmethod
    def SQL_drop_table_parser():
        """
        defines the way to parse the magic command ``%SQL_drop_table``
        """
        parser = MagicCommandParser(prog="SQL_drop_table",
                                    description='drop a table from a database')
        parser.add_argument('table', type=str, help='table', eval_type=str)
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object')
        return parser

    @line_magic
    def SQL_drop_table(self, line):
        """
        defines ``%SQL_drop_table`` which drops a table from a database

        .. nbref::
            :tag: SQL
            :title: SQL_drop_table

            The code for magic command ``%SQL_drop_table`` is equivalent to::

                db.drop_table()

            See notebook :ref:`pyensaesqlmagicrst`.
        """
        parser = self.get_parser(
            MagicSQL.SQL_drop_table_parser, "SQL_drop_table")
        args = self.get_args(line, parser)

        if args is not None:
            db = self.get_connection(args.variable)
            return db.drop_table(args.table)

    @staticmethod
    def SQL_refresh_completion_parser():
        """
        defines the way to parse the magic command ``%SQL_refresh_completion``
        """
        parser = MagicCommandParser(prog="SQL_refresh_completion",
                                    description='refresh completion (tables names, ...)')
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object')
        return parser

    @line_magic
    def SQL_refresh_completion(self, line=""):
        """
        defines ``%SQL_refresh_completion``
        """
        parser = self.get_parser(
            MagicSQL.SQL_refresh_completion_parser, "SQL_refresh_completion")
        args = self.get_args(line, parser)

        if args is not None:
            db = self.get_connection(args.variable)
            db.refresh_completion()

    @staticmethod
    def SQL_schema_parser():
        """
        defines the way to parse the magic command ``%SQL_schema``
        """
        parser = MagicCommandParser(prog="SQL_schema",
                                    description='schema of a table')
        parser.add_argument('table', type=str, help='table', eval_type=str)
        parser.add_argument(
            '--as_list', help='as a dictionary (False) or as a list (True)', action="store_true")
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object')
        return parser

    @line_magic
    def SQL_schema(self, line=""):
        """
        define ``SQL_schema``

        .. nbref::
            :tag: SQL
            :title: SQL_schema

            The code for magic command ``%SQL_schema`` is equivalent to::

                db.get_table_columns(<table>, as_dict=not <as_list>)

            See notebook :ref:`pyensaesqlmagicrst`.
        """
        parser = self.get_parser(MagicSQL.SQL_schema_parser, "SQL_schema")
        args = self.get_args(line, parser)

        if args is not None:
            db = self.get_connection(args.variable)
            return db.get_table_columns(args.table, as_dict=not args.as_list)

    @staticmethod
    def SQL_import_tsv_parser():
        """
        defines the way to parse the magic command ``%SQL_import_tsv``
        """
        parser = MagicCommandParser(prog="SQL_import_tsv",
                                    description='import a tsv file into the database')
        parser.add_argument('filename', type=str,
                            help='tsv file name', eval_type=str)
        parser.add_argument('-t', '--table', type=str,
                            help='table name', default="-", eval_type=str)
        parser.add_argument(
            '--verbose', help='print progress', action="store_true")
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object')
        return parser

    @line_magic
    def SQL_import_tsv(self, line):
        """
        defines ``%SQL_import_tsv`` whichs import a TSV file into a database

        .. nbref::
            :tag: SQL
            :title: SQL_import_tsv

            The code for magic command ``%SQL_import_tsv`` is equivalent to::

                db.import_flat_file(<filename>, <table>)

            See notebook :ref:`pyensaesqlmagicrst`.
        """
        parser = self.get_parser(
            MagicSQL.SQL_import_tsv_parser, "SQL_import_tsv")
        args = self.get_args(line, parser)

        if args is not None:
            if not os.path.exists(args.filename):
                raise FileNotFoundError(args.filename)
            db = self.get_connection(args.variable)
            table = os.path.splitext(os.path.split(args.filename)[-1])[0] \
                if len(args.table) == 0 or args.table == "-" else args.table
            return db.import_flat_file(args.filename, table)

    @staticmethod
    def SQL_add_function_parser():
        """
        defines the way to parse the magic command ``%SQL_add_function``
        """
        parser = MagicCommandParser(prog="SQL_add_function",
                                    description='add a custom function to the database')
        parser.add_argument('funct', type=str, help='function name')
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object')
        return parser

    @line_magic
    def SQL_add_function(self, line):
        """
        defines ``%SQL_add_function`` which adds a function to the database

        .. nbref::
            :tag: SQL
            :title: SQL_add_function

            The code for magic command ``%SQL_add_function`` is equivalent to::

                db.add_function(fu)

            See notebook :ref:`pyensaesqlmagicrst`.
        """
        parser = self.get_parser(
            MagicSQL.SQL_add_function_parser, "SQL_add_function")
        args = self.get_args(line, parser)

        if args is not None:
            db = self.get_connection(args.variable)
            if isinstance(args.funct, str):
                if self.shell is not None:
                    if args.funct not in self.shell.user_ns:
                        raise KeyError(
                            "unable to find function %s in your workspace" %
                            args.funct)
                    fu = self.shell.user_ns[args.funct]
                else:
                    raise Exception("unable to find IPython workspace")
            else:
                fu = args.funct
            return db.add_function(fu)

    @staticmethod
    def SQL_import_df_parser():
        """
        defines the way to parse the magic command ``%SQL_import_df``
        """
        parser = MagicCommandParser(prog="SQL_import_df",
                                    description='import a dataframe into the database')
        parser.add_argument('df', type=str, help='dataframe', no_eval=True)
        parser.add_argument('-t', '--table', type=str,
                            help='table name', default="-", eval_type=str)
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object')
        return parser

    @line_magic
    def SQL_import_df(self, line):
        """
        defines ``%SQL_import_df`` which imports a dataframe into a database

        .. nbref::
            :tag: SQL
            :title: SQL_import_df

            The code for magic command ``%SQL_import_df`` is equivalent to::

                db.import_dataframe(<table>, <df>)

            See notebook :ref:`pyensaesqlmagicrst`.
        """
        parser = self.get_parser(
            MagicSQL.SQL_import_df_parser, "SQL_import_df")
        args = self.get_args(line, parser)

        if args is not None:
            db = self.get_connection(args.variable)

            df = args.df
            if self.shell is not None:
                if df not in self.shell.user_ns:
                    raise KeyError(
                        "unable to find dataframe %s in your workspace" %
                        df)
                odf = self.shell.user_ns[df]
            else:
                raise Exception("unable to find IPython workspace")

            table = df if len(
                args.table) == 0 or args.table == "-" else args.table
            return db.import_dataframe(table, odf)

    @staticmethod
    def SQL_parser():
        """
        defines the way to parse the magic command ``%%SQL``
        """
        parser = MagicCommandParser(prog="SQL",
                                    description='query the database')
        parser.add_argument(
            '--df', type=str, help='output dataframe', default="temp_view", no_eval=True)
        parser.add_argument('-n', '--n', type=int,
                            help='number of first lines to display', default=10, eval_type=int)
        parser.add_argument('-q', '--query', type=str,
                            help='when used in a single line (no cell), query is the SQL query, the command ' +
                            'returns the full dataframe', default="", eval_type=str)
        parser.add_argument(
            '-v',
            '--variable',
            default="DB",
            help='variable name used to store the database object')
        return parser

    @line_cell_magic
    def SQL(self, line, cell=None):
        """
        defines command ``%%SQL``

        .. nbref::
            :tag: SQL
            :title: SQL

            The code for magic command ``%%SQL`` is equivalent to::

                <variable> = db.execute(<cell>)

            See notebook :ref:`pyensaesqlmagicrst`.
        """
        parser = self.get_parser(MagicSQL.SQL_parser, "SQL")
        args = self.get_args(line, parser)

        if args is not None:
            full = False
            if cell is None or len(cell) == 0:
                cell = args.query
                if cell is None or len(cell) == 0:
                    raise ValueError("no SQL query is defined")
                query = cell
                full = True
            else:
                query = cell

            query = query.strip()
            if len(query) > 0 and query[0] == '"' and query[-1] == '"':
                query = query[1:-1]

            db = self.get_connection(args.variable)
            try:
                df = db.execute(query)
                ok = True
            except InterfaceSQLException as e:
                print(str(e))
                ok = False

            if ok:
                self.shell.user_ns[args.df] = df
                if full:
                    return df
                else:
                    return df.head(n=args.n)


def register_sql_magics(ip=None):
    """
    register magics function, can be called from a notebook

    @param      ip      from ``get_ipython()``
    """
    if ip is None:
        from IPython import get_ipython
        ip = get_ipython()
    ip.register_magics(MagicSQL)



"""
@file

@brief @see cl Database
"""
import os
import sys
import math
import re
import time
import decimal
import sqlite3 as SQLite
import datetime
import numpy
from .database_exception import ExceptionSQL, DBException
from .database_core2 import DatabaseCore2

module_odbc = None


class DatabaseCore(DatabaseCore2):
    """
    Core methods for class @see cl Database.

    @var    _engine     engine type (SQLite is the only available)
    @var    _sql_file   database file, if it does not exist, it will be created.
    """

    _sql_keywords = ["order", "by", "select", "from", "group", "where", "as", "like", "upper", "collapse", "join", "union",
                     "inner", "default", "id", "double", "text", "varchar", "float", "long", "Decimal"]

    _SQL_conversion_types = {"": float,
                             "TEXT": str,
                             "text": str,
                             "INTEGER": int,
                             "FLOAT": float,
                             "REAL": float,
                             "float": float,
                             "numeric": float,
                             "LONG": int,
                             "int": int,
                             "varchar": str,
                             "VARCHAR": str,
                             "Decimal": decimal.Decimal,
                             "DATETIME": datetime.datetime,
                             "smallint": int,
                             "bigint": float, }

    _engines = ["SQLite", "MySQL", "ODBCMSSQL"]
    _field_option = ["PRIMARYKEY", "AUTOINCREMENT", "AUTOFILL"]

    def __init__(self, sql_file, engine="SQLite", user=None, password=None,
                 host="localhost", LOG=None, attach=None):
        """
        Creates a database object.

        @param      sql_file        database file database file (use ``:memory:`` to avoid creating a file and using only memory)
                                    it can also contain several files separated by ;
                                    ``name_file ; nickname,second_file ; ...``
        @param      engine          SQLite or MySQL (if it is installed), ODBCMSSQL
        @param      user            user if needed
        @param      host            to connect to a MSSQL database
        @param      password        password if needed
        @param      LOG             LOG function, if None, choose ``print``
        @param      attach          dictionary ``{nickname: filename}``,
                                    list of databases to attach

        @warning If the folder does not exist, it will be created

        Parameter *dbfile* can be of type
        `sqlite3.Connection <https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection>`_.
        """

        # attach cases
        if attach is None:
            attach = {}
        else:
            attach = attach.copy()

        if isinstance(sql_file, str):

            for e in DatabaseCore._engines:
                if sql_file.startswith(e + ":::"):
                    engine = e
                    sql_file = sql_file[len(e) + 3:]
                    if "###" in sql_file:
                        host, sql_file = sql_file.split("###")
                    break

            if ";" in sql_file:
                li = [s.strip() for s in sql_file.split(";")]
                sql_file = li[0]
                rest = li[1:]
                for s in rest:
                    ok = s.split(",")
                    if len(ok) != 2:
                        raise DBException(  # pragma: no cover
                            "unable to find an alias in %r" % s)
                    nick = ok[0].strip()
                    file = ",".join(ok[1:])
                    attach[nick] = file.strip()
            elif sql_file.startswith(":"):
                if sql_file != ":memory:":
                    raise FileNotFoundError(  # pragma: no cover
                        "unable to interpret file: %r" % sql_file)

        # some initialization
        self._password = password
        self._user = user
        self._host = host

        # the rest
        if LOG is None:
            def blind(*li, **p):  # pragma: no cover
                pass
            LOG = blind  # pragma: no cover
        self.LOG = LOG

        if isinstance(LOG, dict):
            raise TypeError(  # pragma: no cover
                "fLOG should be a function, not a dictionary")
        if isinstance(self.LOG, dict):
            raise TypeError(  # pragma: no cover
                "LOG should be a function, not a dictionary")

        if engine == "SQLite":
            self._sql_file = sql_file
            self._engine = engine

        elif engine == "ODBCMSSQL":
            raise DBException(  # pragma: no cover
                "Unable to connect to a SQL server.")

        else:
            raise DBException(  # pragma: no cover
                "unfounded engine %s in %s" %
                (engine, ", ".join(
                    DatabaseCore._engines)))

        # write a file able to build a database summary
        if isinstance(sql_file, str) and not self.isMemory():
            folder = os.path.split(sql_file)[0]
            if len(folder) > 0 and not os.path.exists(folder):
                os.makedirs(folder)
            summary = os.path.join(folder, "temp_quick_look_up.py")
            if not os.path.exists(summary):
                #cwd = os.path.join (os.path.abspath (os.path.split (__file__) [0]), "..", "..")
                #fi  = os.path.split (sql_file) [1]

                if hasattr(DatabaseCore, "SCRIPT_LOOKUP"):
                    script = DatabaseCore.SCRIPT_LOOKUP
                    lines = script.split("\n")
                    lines = [li if "__CWD__ =" not in li else
                             li.replace(
                                 "(__file__)",
                                 "(r'%s')" %
                                 os.path.realpath(__file__))
                             for li in lines]
                    script = "\n".join(lines)
                    script = script.replace(
                        "python quick_look_up.py",
                        "%s quick_look_up.py" %
                        sys.executable)
                    self.LOG("creating script ", summary)
                    try:
                        f = open(summary, "w")
                        f.write(script)
                        f.close()
                    except IOError:
                        self.LOG("unable to write ", summary)

        self._attach = attach
        self._buffer_insert = []
        self._buffer_insert_s = 0

        if isinstance(sql_file, str) and self.isMemory():
            self._connection = SQLite.connect(self._sql_file)
        elif isinstance(sql_file, SQLite.Connection):
            self._connection = sql_file
            self._sql_file = ":memory:"

    def isMSSQL(self):
        """
        Says if the syntax is MS SQL Server.
        """
        if self._engine == "ODBCMSSQL":
            return True
        return False

    def isMemory(self):
        """
        Tells if the Database takes place in memory (``:memory:``).
        """
        return self._sql_file == ":memory:"

    ##########################################################################
    # connection
    ##########################################################################

    def SetBufferInsert(self, n):
        """
        This function offers the possibility to postpone the insertion,
        they will be processed all at the time during when method commit is called.

        @param      n           number of insertion to postpone
        """
        self._buffer_insert_s = n

    def is_connected(self):
        """
        Says if the database is connected.

        @return         "_connection" in self.__dict__
        """
        return "_connection" in self.__dict__

    @staticmethod
    def regex_match(exp, st):
        "Applies a regular expression. Static method to insert in a SQL query."
        return 0 if re.compile(exp).search(st) is None else 1

    @staticmethod
    def idaytodate(dayint, year, month, day):
        "Date conversion. Static method to insert in a SQL query."
        try:
            d = datetime.datetime(year, month, day)
            day = datetime.datetime(year, month, day + 1) - d
            cur = d + day * dayint
            return str(cur).split()[0]
        except Exception as e:
            return str(e)

    @staticmethod
    def isectoday(sec):
        "Date conversion. Static method to insert in a SQL query."
        if sec < 0:
            return "negative time"
        elif sec >= 86400:
            return "out of day"
        else:
            s = int(sec)
            h = s / 3600
            m = (s % 3600) / 60
            s %= 60
            return "%02d:%02d:%02d" % (h, m, s)

    @staticmethod
    def itimestamp(t, year, month, day):
        "Date conversion. Static method to insert in a SQL query."
        d = DatabaseCore.idaytodate(int(t / 86400), year, month, day)
        s = DatabaseCore.isectoday(int(t - 86400. * int(t / 86400)))
        return d + " " + s

    @staticmethod
    def string_to_date(s):
        "Date conversion. Static method to insert in a SQL query."
        d = int(s[:2])
        m = int(s[3:5])
        y = int(s[6:])
        return datetime.datetime(y, m, d)

    @staticmethod
    def _special_function_init_():
        _list_special_function = [
            ("log", math.log, 1, "log(s) --> float", "log"),
            ("exp", math.exp, 1, "exp(s) --> float", "exp"),
            ("len", len, 1, "len(s) --> int", "string length"),
            ("lower",
             lambda s:s.lower(),
             1,
             "lower(s) --> string",
             "lower case"),
            ("upper",
             lambda s:s.upper(),
             1,
             "upper(s) --> string",
             "upper case"),
            ("isubstring", lambda sub, s: 1 if sub in s else 0,
             2, "isubstring(sub,str) --> {0,1}", "return 1 if str includes sub, 0 otherwise"),
            ("match", DatabaseCore.regex_match,
             2, "match(regex,str) --> {0,1}", "return 1 if str matches the regular expression exp, 0 otherwise"),
            ("idaytodate", DatabaseCore.idaytodate,
             4, "idaytodate (day, 1970, 1, 1) --> str", "date if day is the number of days since 01/01/1970"),
            ("itimestamp", DatabaseCore.itimestamp,
             4, "itimestamp (t, 1970, 1, 1) --> str", "date,time if t is the number of seconds since 01/01/1970"),
            ("isectoday", DatabaseCore.isectoday,
             1, "isectoday (isec) --> str", "time if isec is the number of seconds since midnight"),
        ]
        return _list_special_function

    def connect(self, odbc_string=None):
        """
        Opens a connection to the database.

        @param      odbc_string     use a different odbc string
        """
        if self.isMemory():
            if "_connection" not in self.__dict__:
                raise DBException(  # pragma: no cover
                    "It is a database in memory, the database should already be connected.")
        else:
            if "_connection" in self.__dict__:
                raise RuntimeError("A previous connection was not closed.")

            if self._engine == "SQLite":
                self._connection = SQLite.connect(self._sql_file)
            # elif self._engine == "MySQL" :  self._connection =
            # MySQLdb.connect (self._host, self._user, self._password,
            # self._sql_file)
            elif self._engine == "ODBCMSSQL":  # pragma: no cover

                if odbc_string is None:
                    temp = ["DRIVER={SQL Server Native Client 10.0}",  # {SQL Server}",
                            "SERVER=%s" % self._host,
                            "DATABASE=%s" % self._sql_file,
                            "Trusted_Connection=yes",
                            "MARS_Connection=yes",
                            # "MultipleActiveResultSets=True",
                            #"Integrated Security=SSPI",
                            ]
                    #temp = ["DSN=%s" % self._sql_file ]
                    if self._user is not None:
                        temp.append("UID=%s" % self._user)
                    if self._password is not None:
                        temp.append("PASSWORD=%s" % self._password)
                    st = ";".join(temp)
                    self.LOG("connection string ", st)
                    self._connection = module_odbc.connect(st)
                else:
                    st = odbc_string
                    self.LOG("connection string ", st)
                    self._connection = module_odbc.connect(st)

            else:
                raise DBException(  # pragma: no cover
                    "This engine does not exists (%r)" % self._engine)

            for func in DatabaseCore._special_function_init_():
                self.add_function(func[0], func[2], func[1])

            for k, v in self._attach.items():
                self.attach_database(v, k)

    def close(self):
        """
        Closes the database.
        """
        if self.isMemory():
            # we should not close, otherwise, we lose the data
            pass
        else:
            self._check_connection()
            self._connection.close()
            del self._connection

    def commit(self):
        """
        Calls this function after any insert request.
        """
        self._check_connection()

        for s in self._buffer_insert:
            self._connection.execute(s)
        del self._buffer_insert[:]

        self._connection.commit()

    ##########################################################################
    # access part
    ##########################################################################

    def get_file(self, attached_db=False):
        """
        Gets database file.

        @param      attached_db     if True, add the list of attached databases
        @return                     the database file
        """
        if attached_db:
            files = [self._sql_file]
            att = self.get_attached_database_list(True)
            for alias, file in att:
                files.append("%s,%s" % (alias, file))
            temp = ";".join(files)
            if self._engine != "SQLite":
                if self._host is None:
                    temp = "%s:::%s" % (self._engine, temp)
                else:
                    temp = "%s:::%s###%s" % (self._engine, self._host, temp)
            return temp
        else:
            return self._sql_file

    def has_table(self, table):
        """
        Says if the table belongs to the database.

        @param      table       table name
        @return                 boolean
        """
        return table in self.get_table_list("." in table)

    def has_index(self, index):
        """
        Says if the index belongs to the database.

        @param      index       index name
        @return                 boolean"""
        return index in [s[0] for s in self.get_index_list()]

    def get_index_on_table(self, table, full=False):
        """
        Returns the list of indexes on a specific table.

        @param      table       table
        @param      full        if True returns all fields, otherwise, returns only the index names
        @return                 list of the index on this table
        """
        indexes = self.get_index_list()
        if full:
            return [la for la in indexes if la[1] == table]
        return [la[0] for la in indexes if la[1] == table]

    def get_column_type(self, table, column):
        """
        Returns the column type of a table.

        @param      table       table name
        @param      column      column name
        @return                 type (python class)
        """
        self._check_connection()
        cols = self.get_table_columns_list(table)
        for c in cols:
            if c[0] == column:
                return c[1]
        raise DBException(
            "column %s were not found in table %s" %
            (column, table))

    def get_index_list(self, attached="main"):
        """
        Returns the list of indexes.

        @param        attached      if main, returns the index for the main database, otherwise, for an attached database
        @return                     list of tuple (index_name, table, sql_request, fields)
        """
        self._check_connection()
        if attached == "main":
            request = """   SELECT name,tbl_name,sql
                            FROM (SELECT * FROM sqlite_master UNION ALL SELECT * FROM sqlite_temp_master) AS temptbl
                            WHERE type='index' ORDER BY name;"""
        else:
            request = """   SELECT name,tbl_name,sql
                            FROM (SELECT * FROM %s.sqlite_master) AS temptbl
                            WHERE type='index' ORDER BY name;""" % attached
        select = self._connection.execute(request)

        exp = re.compile("[(]([a-zA-Z0-9_,]+)[)]")
        res = []
        for a, b, c in select:
            fi = exp.findall(c)
            if len(fi) != 1:
                raise DBException(  # pragma: no cover
                    "Unable to extract index fields from %r" % c)
            fi = tuple(s.strip() for s in fi[0].split(","))
            res.append((a, b, c, fi))
        select.close()
        #self.LOG ("number of indices ", len (res))
        select = res

        res = []
        if attached == "main":
            res = select
        else:
            for el in select:
                res.append((el[0], attached + "." + el[1], el[2], el[3]))
        #self.LOG ("number of indices ", len (res))

        if attached == "main":
            attach = self.get_attached_database_list()
            for a in attach:
                if a in ("main", "temp"):
                    continue
                r = self.get_index_list(a)
                res.extend(r)

        return res

    def get_attached_database_list(self, file=False):
        """
        Returns all the attached database (avoid the temporary ones and the main one).

        @param      file        ask for file also
        @return                 a list of tuple (alias, file)
        """
        if self.isMSSQL():
            return []  # pragma: no cover
        else:
            cur = self._connection.cursor()
            cur.execute("PRAGMA database_list;")
            res = cur.fetchall()
            cur.close()
            res = [r for r in res if r[1] != "temp" and r[1] != "main"]
            if file:
                return [(r[1], r[2]) for r in res]
            else:
                return [r[1] for r in res]

    def get_table_list(self, add_attached=False):
        """
        Returns the list of tables.

        @param      add_attached        if True, add the list of tables included in the attached databases
        @return                         the table list
        """
        self._check_connection()
        if self.isMSSQL():  # pragma: no cover
            request = """   SELECT TABLE_NAME FROM (
                                SELECT TABLE_NAME, OBJECTPROPERTY(object_id(TABLE_NAME), N'IsUserTable') AS type
                                FROM INFORMATION_SCHEMA.TABLES) AS temp_tbl
                            WHERE type = 1 ORDER BY TABLE_NAME"""
        else:
            request = """   SELECT name
                            FROM (SELECT * FROM sqlite_master UNION ALL SELECT * FROM sqlite_temp_master) AS temptbl
                            WHERE type in('table','temp') AND name != 'sqlite_sequence' ORDER BY name;"""

        select = self._connection.execute(request)
        res = []
        for el in select:
            res.append(el[0])

        if add_attached:
            att = self.get_attached_database_list()
            for at in att:
                if at == "temp":
                    continue
                sql = "SELECT name FROM %s.sqlite_master" % at
                vie = self._connection.execute(sql)
                vie = ["%s.%s" % (at, v[0]) for v in vie]
                res.extend(vie)
        return res

    def get_table_columns(self, table, dictionary=False):
        """
        See @see me get_table_columns_list.

        Example (`dictionary == False`):
        ::
            [('fid', <type 'int'>), ('fidp', <type 'int'>), ('field', <type 'str'>)]

        Or (`dictionary = True`):
        ::
            {0: ('fid', <type 'int'>), 1: ('fidp', <type 'int'>), 2: ('field', <type 'str'>)}
        """
        return self.get_table_columns_list(table, dictionary)

    def get_table_columns_list(self, table, dictionary=False):
        """
        Returns all the columns for a table.

        @param      table       table name
        @param      dictionary  returns the list as a dictionary
        @return                 a list of tuple (column name, Python type)

        Example (`dictionary == False`):

        ::
            [('fid', <type 'int'>), ('fidp', <type 'int'>), ('field', <type 'str'>)]

        Or (`dictionary = True`):

        ::

            {0: ('fid', <type 'int'>), 1: ('fidp', <type 'int'>), 2: ('field', <type 'str'>)}
        """
        if "." in table:
            prefix = table.split(".")[0] + "."
            table = table.split(".")[1]
        else:
            # table = table
            prefix = ""
        cur = self._connection.cursor()

        if self.isMSSQL():  # pragma: no cover
            prf = "" if len(prefix) == 0 else prefix + "."
            sql = """SELECT * FROM (SELECT OBJECT_NAME(c.OBJECT_ID) TableName,c.name AS ColumnName,t.name AS TypeName
                            FROM sys.columns AS c
                            JOIN sys.types AS t ON c.user_type_id=t.user_type_id
                            ) AS ttt
                            WHERE ttt.TableName = '%s%s'""" % (prf, table)
            cur.execute(sql)
        else:
            cur.execute("PRAGMA %stable_info(%s)" % (prefix, table) + ";")

        res = cur.fetchall()
        cur.close()
        res = [(r[1], DatabaseCore._SQL_conversion_types[r[2]]) for r in res]
        if dictionary:
            dic = {}
            for i in range(0, len(res)):
                dic[i] = res[i]
            return dic
        else:
            return res

    def get_table_nb_lines(self, table):
        """
        Returns the number of lines in a table (or number of observations).

        @param      table       table name
        @return                 integer
        """
        sql = "SELECT COUNT(*) FROM " + table + ";"
        cur = self._connection.cursor()
        cur.execute(sql)
        res = cur.fetchall()
        cur.close()
        return res[0][0]

    def len(self, table):
        """
        Returns the number of lines of table ``table``.

        @param  table       table
        @return             int
        """
        return self.get_table_nb_lines(table)

    def get_table_nfirst_lines(self, table, n=1):
        """
        Returns the *n* first lines.

        @param      table       table name
        @param      n           number of asked lines
        @return                 integer
        """
        sql = "SELECT * FROM %s ;" % table
        cur = self._connection.cursor()
        cur.execute(sql)
        if n <= 1:
            res = [cur.fetchone()]
        else:
            res = []
            for line in cur:
                n -= 1
                if n <= -1:
                    break
                res.append(line)
        cur.close()
        return res

    def get_sql_columns(self, request):
        """
        Returns the columns name for a SQL request.

        @param      request             SQL request
        @return                         list of columns name
        """
        cur = self.execute(request)
        col_name_list = [tuple[0] for tuple in cur.description]
        cur.close()
        return col_name_list

    ##########################################################################
    # execution
    ##########################################################################

    class _cross_product_iter:

        """
        Iterator for CROSS.
        """

        def __init__(self, db, request):
            com = re.compile("^(.*)(--.*)$")
            lines = request.split("\n")
            clean = []
            for li in lines:
                r = com.match(li)
                if r is not None:
                    li = li[:r.span(1)[1]]
                clean.append(li.strip())
            req = " ".join(clean)
            cross = re.compile(" *CROSS +([ a-zA-Z_0-9,]+?) +"
                               "PLACE +([,a-zA-Z_0-9()]+?) +"
                               "FROM +([a-zA-Z_0-9]+?)( +AS +[_a-z0-9])? +ORDER +BY "
                               "+([ a-zA-Z_0-9,]+?)( +WHERE(.*?))?( +LIMIT +([0-9]+)?)?$", re.IGNORECASE)
            db.LOG("cross product", req)
            find = cross.search(req)
            self.request = request
            self.find = find
            self.db = db

            if self.find is None:
                return

            gr = self.find.groups()
            key, value, table, count_as, order, where, _, __, limit = gr
            if limit is not None:
                limit = int(limit)
            db.LOG("parameters ", [key, value, table, order, where, limit])
            fkey = key.split(",")
            fval = value.split(",")
            if where is None:
                where = ""

            nkey = len(fkey)
            #nval = len (fval)
            sql = "SELECT %s,%s FROM %s %s ORDER BY %s" % (
                key, value, table, where, order)
            cur = self.db.execute(sql)
            data = {}
            #tot  = nkey + nval
            for sample in cur:
                key = sample[:nkey]
                val = sample[nkey:]
                if key not in data:
                    data[key] = []
                data[key].append(val)
            cur.close()

            keys = sorted(data.keys())

            if nkey == 1:
                temp = [[str(k[0]) + ';' + s for s in fval] for k in keys]
            else:
                temp = [
                    [",".join([str(_s) for _s in k]) + ';' + s for s in fval] for k in keys]
            self.description = []
            for t in temp:
                self.description.extend(t)
            self.description = [(k, None) for k in self.description]

            matrix = []
            pos = 0
            while True:
                stil = 0
                line = []
                for k in keys:
                    v = data[k]
                    if pos < len(v):
                        stil += 1
                        line.extend(list(v[pos]))
                    else:
                        line.extend([None for f in fval])
                if stil > 0:
                    matrix.append(line)
                    pos += 1
                    if limit is not None and pos >= limit:
                        break
                else:
                    break
            self.matrix = matrix
            self.pos = 0

        def is_working(self):
            return self.find is not None

        def __iter__(self):
            """
            iterator
            """
            return self

        def __next__(self):
            """
            iterator
            """
            if "matrix" not in self.__dict__:
                raise StopIteration  # pragma: no cover
            if self.pos < len(self.matrix):
                n = self.pos
                self.pos += 1
                return self.matrix[n]
            else:
                raise StopIteration

        def close(self):
            pass

    def _analyse(self, request, header=False):
        """
        Analyses the request does it contains cross product.

        @param      request     request
        @param      header      add a header in the first line
        @return                 None or an iterator

        Example:

        ::

            CROSS f1,f2,f3
            PLACE a,b,c
            FROM table
            ORDER BY f8
            WHERE f9 == ' '  -- optional
        """
        if "CROSS" not in request.upper():
            return None
        iter = DatabaseCore._cross_product_iter(self, request)
        if not iter.is_working():
            return None
        else:
            return iter

    def execute(self, request, nolog=False):
        """
        Opens a cursor with a query and return it to the user.

        @param      request         SQL request
        @param      nolog           if True, do not log anything
        @return                     cursor

        .. exref::
            :title: run a select command on a table
            :tag: SQL

            ::

                t = Database (file)
                cur = t.execute ("SELECT * FROM table1 ;")
                for f in cur :
                    print(f)
                cur.close ()

        There is another case outside SQL syntax to build cross product. Syntax:

        ::

            CROSS f1,f2,f3
            FROM table
            PLACE a,b,c
            ORDER BY f8
            WHERE f9 == ' '  -- optional

        The request must begin by CROSS
        """
        res = self._analyse(request)
        if res is not None:
            return res
        else:
            # classic ways
            self._check_connection()
            cur = self._connection.cursor()
            dat = time.perf_counter()
            try:
                if not nolog:
                    lines = request.split("\n")
                    if len(lines) > 20:
                        self.LOG("SQL ", "\n".join(
                            [repr(x) for x in lines[:20]]))
                    else:
                        self.LOG("SQL ", "\n".join([repr(x) for x in lines]))
                cur.execute(request)
                dat2 = time.perf_counter()
                if dat2 - dat > 10:
                    self.LOG("SQL end")  # pragma: no cover
            except Exception as e:
                raise ExceptionSQL(
                    "unable to execute a SQL request (1)(file %s)" %
                    self.get_file(),
                    e,
                    request) from e
            return cur

    def execute_view(self, request, add_column_name=False, nolog=True):
        """
        Opens a cursor with a query and returns the result into a list.

        @param      request             SQL request
        @param      add_column_name     add the column name before the first line
        @param      nolog               if True, do not log anything
        @return                         cursor

        Example:

        ::

            t = Database (file)
            view = t.execute_view ("SELECT * FROM table1 ;")
        """
        cur = self.execute(request, nolog=nolog)
        if add_column_name:
            col_name_list = [tuple[0] for tuple in cur.description]
            res = [col_name_list] + list(cur)
        else:
            res = list(cur)
        cur.close()
        if not nolog and (len(res) == 0 or len(res) > 1e4):
            self.LOG("execute_view ", len(res), "results")  # pragma: no cover
        return res

    def execute_script(self, script, nolog=True, close=True):
        """
        Opens a cursor and run a script.

        @param      script              SQL script
        @param      nolog               if True, do not log anything
        @param      close               close the cursor
        @return                         cursor
        """
        self._check_connection()
        if not nolog:  # pragma: no cover
            lines = script.split("\n")
            if len(lines) > 20:
                self.LOG("SQL start + ",
                         "\n".join([repr(x) for x in lines[:20]]))
            else:
                self.LOG("SQL start + ",
                         "\n".join([repr(x) for x in lines]))
        cur = self._connection.cursor()
        res = cur.executescript(script)
        if close:
            cur.close()
            if not nolog:
                self.LOG("SQL end")  # pragma: no cover
        else:
            return res

    ##########################################################################
    # extra functions
    ##########################################################################

    def attach_database(self, db, alias):
        """
        Attaches another database.

        @param      db          database to attach
        @param      alias       database alias
        """
        if isinstance(db, str):
            self.LOG("ATTACH DATABASE '%s' TO '%s' ALIAS %s" % (db, db, alias))
            self.execute("ATTACH DATABASE '%s' AS %s;" % (db, alias))
        else:  # pragma: no cover
            self.LOG(
                "ATTACH DATABASE '%s' TO '%s' ALIAS %s" %
                (db._sql_file, self._sql_file, alias))
            self.execute(
                "ATTACH DATABASE '%s' AS %s;" %
                (db.get_file(), alias))

    def add_function(self, name, nbparam, function):
        """
        Adds a function which can be used as any other SQL function (strim, ...).

        @param      name            function name (it does not allow _)
        @param      nbparam         number of parameters
        @param      function        function to add
        """
        if "_" in name:
            raise RuntimeError(  # pragma: no cover
                "SQLite does not allow function name with _")
        self._check_connection()
        if self._engine == "SQLite":
            self._connection.create_function(name, nbparam, function)

    ##########################################################################
    # creation function
    ##########################################################################

    def create_index(self, indexname, table, columns, unique=False):
        """
        Creates an index on a table using some columns.

        @param      indexname   index name
        @param      table       table name
        @param      columns     list of columns
        @param      unique      any value in the columns is unique?
        """
        if not isinstance(columns, list) and not isinstance(columns, tuple):
            columns = [columns]

        if "." in table:
            prefix = table.split(".")[0] + "."
            table = table.split(".")[1]
        else:
            prefix = ""
            # table = table

        self.LOG("index create ", indexname, table, columns, unique)
        if unique:
            sql = "CREATE UNIQUE INDEX %s%s ON %s (%s);" % (
                prefix, indexname, table, ",".join(columns))
        else:
            sql = "CREATE INDEX %s%s ON %s (%s);" % (
                prefix, indexname, table, ",".join(columns))
        self.execute(sql)

    def create_table(self, table, columns, temporary=False, nolog=False):
        """
        Creates a table.

        @param      table           table name
        @param      columns         columns definition, dictionary { key:(column_name,python_type) }
                                    if PRIMARYKEY is added, the key is considered as the primary key.
        @param      temporary       if True the table is temporary
        @param      nolog           @see me execute
        @return                     cursor

        Example for *columns*:

        ::

            columns = { -1:("key", int, "PRIMARYKEY", "AUTOINCREMENT"),
                         0:("name",str), 1:("number", float) }

        """
        if self._engine == "SQLite" and table == "sqlite_sequence":
            raise DBException(  # pragma: no cover
                "unable to create a table named sql_sequence")

        tables = self.get_table_list()
        if table in tables:
            raise DBException(  # pragma: no cover
                "table %r is already present, it cannot be added" % table)

        if temporary:
            sql = "CREATE TEMPORARY TABLE " + table + "("
        else:
            sql = "CREATE TABLE " + table + "("
        col = []
        for c, val in columns.items():
            if self.isMSSQL():  # pragma: no cover
                if isinstance(val[1], tuple):
                    v, li = val[1]
                else:
                    v, li = val[1], 2048

                if li > 8000:
                    col.append(val[0] + " TEXT")
                elif v is str:
                    col.append(val[0] + " VARCHAR(%d)" % li)
                elif v is int:
                    col.append(val[0] + " INTEGER")
                elif v is float:
                    col.append(val[0] + " FLOAT")
                elif v is numpy.int64:
                    col.append(val[0] + " INTEGER")
                elif v is numpy.float64:
                    col.append(val[0] + " FLOAT")
                elif v is decimal.Decimal:
                    col.append(val[0] + " Decimal")
                elif v is datetime.datetime:
                    col.append(val[0] + " DATETIME")
                else:
                    raise DBException(  # pragma: no cover
                        "unable to add column " +
                        str(c) +
                        " ... " +
                        str(val) +
                        " v= " +
                        str(v))
            else:
                if isinstance(val[1], tuple):
                    v, li = val[1]
                else:
                    v, li = val[1], 2048

                if v is str:
                    col.append(val[0] + " TEXT")
                elif v is int:
                    col.append(val[0] + " INTEGER")
                elif v is float:
                    col.append(val[0] + " FLOAT")
                elif v is numpy.int64:
                    col.append(val[0] + " INTEGER")
                elif v is numpy.float64:
                    col.append(val[0] + " FLOAT")
                elif v is decimal.Decimal:
                    col.append(val[0] + " Decimal")
                elif v is datetime.datetime:
                    col.append(val[0] + " DATETIME")
                else:
                    raise DBException(  # pragma: no cover
                        "unable to add column " +
                        str(c) +
                        " ... " +
                        str(val) +
                        " v= " +
                        str(v))

            fval = val[2:]
            for v in fval:
                if v not in DatabaseCore._field_option:
                    raise DBException(  # pragma: no cover
                        "an option is unexpected %s should be in %s" %
                        (v, str(
                            DatabaseCore._field_option)))

            if "PRIMARYKEY" in val:
                if val[1] != int:
                    raise DBException(
                        "unable to create a primary key on something differont from an integer (%s)" %
                        str(val))
                col[-1] += " PRIMARY KEY"
                if "AUTOINCREMENT" in val:
                    if self.isMSSQL():
                        col[-1] += " IDENTITY(0,1)"
                    else:
                        col[-1] += " AUTOINCREMENT"

        sql += ",\n       ".join(col)
        sql += ");"
        return self.execute(sql, nolog=nolog)

    ##########################################################################
    # deletion
    ##########################################################################

    def remove_table(self, table):
        """
        Removes a table.

        @param      table       table name
        @return                 return a cursor
        """
        self.execute("DROP TABLE %s" % table)

    ##########################################################################
    # modification
    ##########################################################################

    def _insert_sql(self, table, insert_values):
        """
        Builds the sql for an insert request.

        @param      table               table name
        @param      insert_values       dictionary or a list
        @return                         string
        """
        if isinstance(insert_values, dict):
            keys = []
            values = []
            for k, v in insert_values.items():
                keys.append(k)
                if v is None:
                    values.append('')
                elif isinstance(v, str):
                    v = "'" + str(v).replace("'", "''") + "'"
                    values.append(v)
                elif isinstance(v, datetime.datetime):
                    v = "'" + str(v) + "'"
                    values.append(v)
                else:
                    values.append(str(v))
            keys = ",".join(keys)
            values = ",".join(values)
            sql = "INSERT INTO %s (%s) VALUES (%s)" % (table, keys, values)
            return sql
        elif isinstance(insert_values, (tuple, list)):
            values = []
            for v in insert_values:
                if v is None:
                    values.append('')
                elif isinstance(v, str):
                    v = "'" + str(v).replace("'", "''") + "'"
                    values.append(v)
                elif isinstance(v, datetime.datetime):
                    v = "'" + str(v) + "'"
                    values.append(v)
                else:
                    values.append(str(v))
            values = ",".join(values)
            sql = "INSERT INTO %s VALUES (%s)" % (table, values)
            return sql
        else:
            raise TypeError(  # pragma: no cover
                "unexpected type: " + str(type(insert_values)))

    def insert(self, table, insert_values, cursor=None, nolog=True):
        """
        Inserts into a table.

        @param      table           table name
        @param      insert_values   values to insert (a list of dictionary or a single dictionary)
        @param      cursor          if *cursor is not None*, use it, otherwise creates a new one
        @param      nolog           if True, do not log anything
        @return                     sql request or None if several insertion were sent (result is too long)

        @warning The commit is not done and must be done to stored these modifications.
        """
        if isinstance(insert_values, list):
            # we expect several insertion
            if self._engine != "SQLite":
                for d in insert_values:
                    self.insert(table, d, cursor)
            else:
                if isinstance(insert_values[0], dict):
                    ins = {}
                    for k in insert_values[0]:
                        ins[k] = ":" + k
                    sql = self._insert_sql(table, ins)
                else:
                    q = tuple('?' for _ in insert_values[0])
                    sql = self._insert_sql(table, q).replace("'", "")

                sql = sql.replace("'", "")
                try:
                    if not nolog:  # pragma: no cover
                        if len(sql) > 1000:
                            self.LOG("SQLs", sql[:1000])
                        else:
                            self.LOG("SQLs", sql)
                    self._connection.executemany(sql, insert_values)
                    return ""
                except Exception as e:
                    raise ExceptionSQL(  # pylint: disable=W0707
                        "Unable to execute a SQL request (3) (cursor %r) (file %r)" %
                        (str(cursor), self.get_file()), e, sql)

        elif isinstance(insert_values, dict):
            sql = self._insert_sql(table, insert_values)

            try:
                if not nolog:  # pragma: no cover
                    if len(sql) > 1000:
                        self.LOG("SQLs", sql[:1000])
                    else:
                        self.LOG("SQLs", sql)
                if cursor is not None:
                    cursor.execute(sql)
                else:
                    if self._buffer_insert_s > 0:
                        self._buffer_insert.append(sql)

                        if len(self._buffer_insert) >= self._buffer_insert_s:
                            for s in self._buffer_insert:
                                self._connection.execute(s)
                            del self._buffer_insert[:]
                    else:
                        self._connection.execute(sql)

                return sql
            except Exception as e:
                raise ExceptionSQL(  # pylint: disable=W0707
                    "unable to execute a SQL request (2) (cursor %r) (file %r)" %
                    (str(cursor), self.get_file()), e, sql)

        else:
            raise DBException(  # pragma: no cover
                "insert: expected type (list of dict or dict) instead of %s" %
                (str(
                    type(insert_values))))

    def update(self, table, key, value, values):
        """
        Updates some values ``WHERE key=value``.

        @param      table       table to update
        @param      key         key
        @param      value       WHERE key = value
        @param      values      values to be updated

        @warning The commit is not done and must be done
                 to stored these modifications.
        """

        self._check_values(values)
        self._check_connection()
        alls = []
        for k, v in values.items():
            if k != key:
                if isinstance(v, (str, datetime.datetime)):
                    alls += ["%s='%s'" % (k, str(v))]
                else:
                    alls += ["%s=%s" % (k, str(v))]
        if isinstance(value, str):
            sql = "UPDATE %s SET %s WHERE %s='%s'" % (
                table, ",".join(alls), key, value.replace("'", "''"))
        elif isinstance(value, datetime.datetime):
            sql = "UPDATE %s SET %s WHERE %s='%s'" % (
                table, ",".join(alls), key, str(value))
        else:
            sql = "UPDATE %s SET %s WHERE %s=%s" % (
                table, ",".join(alls), key, value)

        try:
            self._connection.execute(sql)
            return sql
        except Exception as e:  # pragma: no cover
            raise ExceptionSQL(  # pylint: disable=W0707
                "Unable to execute a SQL request (4) (file %r)" %
                self.get_file(), e, sql)



"""
@file

@brief contains a class which opens a text file as a binary file.
"""


import re
import os
import math
import time
import decimal
from pyquickhelper.loghelper import noLOG
from .type_helpers import guess_type_value


class TextFile:

    """
    This class opens a text file as if it were a binary file.
    It can deal with null characters which are missed by open function.

    @var    filename        file name
    @var    errors          decoding in utf8 can raise some errors,
                            see `str <https://docs.python.org/3.4/library/stdtypes.html?highlight=str#str>`_
                            to understand the meaning of this parameter
    @var    LOG             logging function
    @var    _buffer_size    read a text file _buffer_size bytes each time
    @var    _filter         function filter, None or return True or False whether a line should considered or not
    @var    _encoding       encoding

    Example:

    ::

        f = TextFile(filename)
        f.open ()
        for line in f :
            print line
        f.close ()
    """

    _split_expr = re.compile("\\r?\\t", re.U)
    _sep_available = "\t;,| "

    def __init__(self, filename, errors=None, fLOG=noLOG, buffer_size=2 ** 20,
                 filter=None, separated=False, encoding="utf-8"):
        """
        @param      filename        filename
        @param      errors          see str (errors = ...)
        @param      fLOG            LOG function, see `fLOG <http://www.xavierdupre.fr/app/pyquickhelper/
                                    helpsphinx/pyquickhelper/loghelper/flog.html#pyquickhelper.loghelper.flog.fLOG>`_
        @param      buffer_size     buffer_size (mostly use to test the reading function)
        @param      filter          None if there is no filter, otherwise it is a function which takes a list and returns a boolean
                                    which tells if the line must considered or not
        @param      separated       if True, the line returned by the iterator are splitted by the most probable separator
        """
        self.filename = filename
        self._encoding = encoding
        self.errors = errors
        self.LOG = fLOG
        self._buffer_size = buffer_size
        self._filter = filter
        self._separated = separated

    def open(self):
        """
        Opens the file in reading mode.
        """
        self.LOG("  TextFile: opening file ", self.filename)
        if self._separated:
            res = self.guess_columns()
            self.LOG("   TextFile: guessed: ", res)
            sep = res[2]
            self._separated_value = sep

        self._f = open(self.filename, "r", encoding=self._encoding)
        self._nbline = 0
        self._read = 0

    def close(self):
        """
        Closes the file.
        """
        self._f.close()
        self.LOG("  TextFile: closing file ", self.filename)
        del self.__dict__["_f"]

    def get_nb_readlines(self):
        """
        Returns the number of read lines.
        """
        return self._nbline

    def get_nb_readbytes(self):
        """
        Returns the number of read bytes.
        """
        return self._nbline

    def readlines(self):
        """
        Extracts all the lines,
        the file must not be opened through method open
        ``\\n`` are removed.
        """
        self.open()
        res = []
        for line in self:
            li = line.strip("\r\n")
            res.append(li)
        self.close()
        return res

    def __iter__(self):
        """
        Iterator

        ::

            f = open('...', 'r')
            for line in f :
                ...
            f.close ()

        @return         a str string
        """
        if "_f" not in self.__dict__:
            raise Exception("file %s is not opened." % self.filename)
        filesize = os.stat(self.filename).st_size
        size = self._buffer_size
        blin = self._f.read(size)
        self._read = len(blin)

        if blin.startswith("\xef\xbb\xbf"):
            self.LOG("  removing the three first character u'\\xef\\xbb\\xbf'")
            blin = blin[3:]
        if blin.startswith("\ufeff"):
            self.LOG("  removing the three first character u'\\ufeff'")
            blin = blin[len("\ufeff"):]

        endline = "\n"
        endchar = "\r"

        begin = 0
        sel = 0
        tim = time.perf_counter()
        while len(blin) > 0:

            pos = blin.find(endline, begin)
            if pos == -1:
                pos = blin.find(endchar, begin)
                if begin == 0 and pos != -1:
                    self.LOG("  problem in file ", self.filename)
                    self.LOG("  the line separator is not \\n but \\r")

            while pos == -1:
                if begin > 0:
                    blin = blin[begin:]
                    begin = 0
                temp = self._f.read(size)
                self._read += len(temp)
                blin += temp
                pos = blin.find("\n")
                if pos == -1:
                    pos = blin.find("\r", begin)
                if len(temp) == 0 and pos == -1:
                    pos = len(blin)

            temp = blin[begin:pos]
            line = str(temp)
            begin = pos + 1

            tim2 = time.perf_counter()
            if tim2 - tim > 60:
                tim = tim2
                ratio = float(self._read) / filesize * 100
                self.LOG(
                    "  processing line ",
                    self._nbline,
                    " read bytes ",
                    self._read,
                    " sel ",
                    sel,
                    " ratio %2.2f" %
                    ratio,
                    "%")

            r = line
            if self._encoding == "utf-8":
                r = r.rstrip(endchar)
            if self._filter is None or self._filter(r):
                if self._separated:
                    yield r.split(self._separated_value)
                else:
                    yield r

            self._nbline += 1

    def _load(self, filename, this_column, file_column, prefix, **param):
        """
        load...
        """
        f = TextFile(filename, fLOG=self.LOG, encoding=self._encoding, **param)
        f.open()
        cont = {}
        for line in f:
            if f.get_nb_readlines() == 0:
                columns = self._interpret_columns(line)
            else:
                col = self._interpret(line)
                key = col[columns[file_column]]
                cont[key] = col
        f.close()
        return cont, columns, this_column, file_column, prefix

    def _interpret_columns(self, line):
        """
        Interprets the first line which contains the columns name.

        @param      line        string
        @return                 dictionary { name:position }"""
        col = self._interpret(line)
        res = {}
        for i in range(0, len(col)):
            res[col[i]] = i
        return res

    def _interpret(self, line):
        """
        Splits a line into a list, separator ``\\t``.

        @param      line        string
        @return                 list
        """
        col = TextFile._split_expr.split(line.strip(" \r\n"))
        return col

    def join(self, definition, output, missing_value="", unique=None, **param):
        """
        Joins several files together.

        @param  definition      list of triplets:
                                    filename, this_column, file_column, prefix
        @param  output          if None, return the results as a list, otherwise save it into output
        @param  param           parameter used to open files
        @param  missing_value   specify a value for the missing values
        @param  unique          if unique is a column name, do not process a line whose value has already been processed, None otherwise
        @return                 columns, matrix or number of of missing values

        We assume that every file starts with header giving columns names.
        The function associates *this_column* value to *file_column* and
        appends all the columns from filename with a prefix.
        We also assumes values in file_column are unique.
        """
        if output is not None:
            output = open(output, "w", encoding=self._encoding)

        files = []
        for i, tu in enumerate(definition):
            if len(tu) == 2:
                a, b = tu
                c = b
                d = "f%d_" % (i + 1)
            elif len(tu) == 3:
                a, b, c = tu
                d = "f%d_" % (i + 1)
            elif len(tu) == 4:
                a, b, c, d = tu
            else:
                raise ValueError(
                    "definition must contain tuple (size, 2, 3 ,4), not {0}".format(tu))
            files.append(self._load(a, b, c, d, **param))

        res = []
        miss = 0
        uniquekey = {}

        self.open()
        for line in self:
            if self.get_nb_readlines() == 0:
                columns = self._interpret_columns(line)
                oldnb = len(columns)
                last = max(columns.values()) + 1
                for file in files:
                    col = file[1]
                    pre = file[-1]
                    for k, v in col.items():
                        columns[pre + k] = last + v
                    last += len(col)
                linecol = ["" for c in columns]
                for k, v in columns.items():
                    linecol[v] = k

                if output is None:
                    res.append(linecol)
                else:
                    output.write("\t".join(linecol) + "\n")

                s1 = len(linecol)
                s2 = oldnb
                for f in files:
                    s2 += len(f[1])
                if s1 != s2:
                    mes = "size problem %d != " % (s1)
                    mes += " + ".join([str(x)
                                       for x in [oldnb, ] + [len(f[1]) for f in files]])
                    raise Exception(mes)

            else:
                col = self._interpret(line)

                if unique is not None:
                    key = columns[unique]
                    val = col[key]
                    if val in uniquekey:
                        uniquekey[val] += 1
                        continue
                    uniquekey[val] = 1

                if len(col) != oldnb:
                    col.extend(["" for i in range(0, oldnb - len(col))])
                if len(col) != oldnb:
                    mes = "line %d: problem len(col) = %d and oldnb = %d\n%s" % (
                        self.get_nb_readlines(), len(col), oldnb, repr(line))
                    raise Exception(mes)

                for file in files:
                    cont = file[0]
                    c = file[1]
                    this_key = col[columns[file[2]]]
                    if this_key in cont:
                        val = cont[this_key]
                        if len(val) == 0 or (len(val) == 1 and len(val[0]) == 0):
                            # empty line
                            continue
                        if len(val) != len(c):
                            ll = self.get_nb_readlines()
                            mes = "line %d: problem len(val) = %d and len (c) = %d\n\"%s\"" % (
                                ll, len(val), len(c), file)
                            raise Exception(mes)
                    else:
                        val = [missing_value for k in c]
                        miss += len(val)
                    col.extend(val)

                if len(col) != len(columns):
                    vals = list(set(col))
                    if vals == ['']:
                        continue
                    mes = "problem 1 with line %d\n" % self.get_nb_readlines()
                    mes += "len (col) = %d len (columns) = %d" % (len(col),
                                                                  len(columns))
                    raise Exception(mes)

                if len(("\t".join(col)).split("\t")) != len(col):
                    mes = "problem 2 with line %d\n" % self.get_nb_readlines()
                    mes += "len (col) = %d len (columns) = %d" % (
                        len(("\t".join(col)).split("\t")), len(columns))
                    raise Exception(mes)

                if output is None:
                    res.append(col)
                else:
                    output.write("\t".join(col) + "\n")

        if output is None:
            return res
        else:
            output.close()
            return miss

    def _count_s(self, car):
        """
        Returns the number of every character in car.
        """
        res = {}
        for i, c in enumerate(car):
            if c in res:
                res[c] += 1
            else:
                res[c] = 1
        return res

    def _get_type(self, s):
        """
        Guesses the type of value s.
        """
        return guess_type_value(s)

    def guess_columns(self, nb=100, force_header=False, changes=None, force_noheader=False,
                      fields=None, regex=None, force_sep=None, mistake=3):
        """
        Guesses the columns type.

        @param      nb              number of lines to have a look to in order to find all the necessary elements
        @param      force_header    impose a header whether it is detect or not
        @param      changes         modify some column names, example { "query":"query___" }
        @param      force_noheader  there is no header at all
        @param      fields          name of the columns if there is no header (instead of c000, c001...)
        @param      regex           if the default expression for a field is not the expected one, change by looking into regex
        @param      force_sep       force the separator to be the one chosen by the user (None by default)
        @param      mistake         not more than mistake conversion in numbers are allowed
        @return                     4-tuple, see below

        Returned result is a 4 t-uple:

        - True or False: presence of a header (it means
          there is at least one numerical column)
        - column definition ``{ position : (name, type) }`` or
          ``{ position : (name, (str, max_length*2)) }``
        - separator
        - regex which allow the user to extract information from the file

        The column separator is looked into ``, | ; \\t``
        @warning The file must not be opened, it will be several times.
        """
        if changes is None:
            changes = {}
        if regex is None:
            regex = {}
        self.LOG("  TextFile.guess_columns: processing file ", self.filename)

        endlinechar = "\n "

        # n lines
        temp = TextFile(self.filename, encoding=self._encoding, fLOG=self.LOG)
        lines = []

        temp.open()
        for line in temp:
            line = line.strip(endlinechar)
            if len(line) == 0:
                continue
            lines.append(line)
            if len(lines) > nb:
                break
        self.LOG("  TextFile.guess_columns: using ", len(lines), " lines")
        temp.close()

        # guess the separation
        sep = TextFile._sep_available
        if force_sep not in (None, force_sep):
            sep += force_sep
        h = {}
        mx = 0
        for line in lines:
            co = self._count_s(line)
            for s in sep:
                n = co.get(s, 0)
                if n == 0:
                    continue
                k = s, n
                if k not in h:
                    h[k] = 1
                else:
                    h[k] += 1
                mx = max(n, mx)

        mx += 1
        best = None
        iner = None
        for c in sep:
            m = {}
            z = 0
            for k in range(mx):
                if (c, k) in h:
                    m[k] = h[c, k]
                    z += k * m[k]

            if len(m) == 0:
                continue
            g = max(sum(m.values()), len(lines))
            if z < max(len(lines) * 9 / 10, 1):
                continue

            for k in m:
                m[k] = float(m[k]) / g
            s = 0.0
            for k in m:
                s += m[k] * math.log(m[k])
            if iner is None or s > iner:
                iner = s
                best = c

        bestsep = best

        if force_sep is not None and bestsep != force_sep:
            self.LOG(
                "  TextFile.guess_columns: changes the separator",
                repr(force_sep))
            bestsep = force_sep

        bestcol = 0
        bestnb = 0
        for k in range(mx):
            if (bestsep, k) in h:
                if bestnb < h[bestsep, k]:
                    bestnb = h[bestsep, k]
                    bestcol = k + 1

        self.LOG("  TextFile.guess_columns: sep ", repr(bestsep), "nb cols", bestcol, " bestnb ",
                 bestnb, " more ", h)

        # determine the type of every column

        h = {}
        for line in lines:
            cols = line.split(bestsep)
            for i in range(len(cols)):
                ty = self._get_type(cols[i])
                k = i, ty
                if k not in h:
                    h[k] = 1
                else:
                    h[k] += 1

        columns = {}
        for a in h:
            k, t = a
            if k >= bestcol:
                continue
            if k not in columns:
                columns[k] = (t, h[a])
            elif h[a] > columns[k][1]:
                columns[k] = (t, h[a])

        for pos in columns:
            # int and float corrections
            if columns[pos][0] == int and h.get((pos, float), 0) > 0:
                self.LOG(
                    "  changing column type ",
                    pos,
                    columns[pos],
                    " into ",
                    float)
                columns[pos] = (float, h[pos, float] + h[pos, int])
            su = h.get((pos, str), 0)
            if (columns[pos][0] == int or columns[pos][0] == float or columns[
                    pos][0] == decimal.Decimal) and su > mistake:
                self.LOG(
                    "  changing column type ",
                    pos,
                    columns[pos],
                    " into ",
                    str,
                    " mistakes ",
                    su,
                    " > ",
                    mistake)
                columns[pos] = (str, columns[pos][1] + su)

        # header or not

        mat = 0
        no = 0
        cols = lines[0].split(bestsep)
        for i, c in enumerate(cols):
            t = self._get_type(c)
            e = columns.get(i, (str, 0))[0]
            if e != str:
                if t == e:
                    mat += 1
                else:
                    no += 1
        header = not force_noheader and (force_header or (no > mat))

        # determine the column name

        if header:
            names = lines[0].split(bestsep)
            del lines[0]
            if len(names) != bestcol:
                raise Exception(
                    "unable to continue: the header does not contain the same number of columns %s != %s" %
                    (len(names), bestcol))
        elif fields is not None:
            if len(fields) != bestcol:
                raise Exception(
                    "the number of fields (%d) is different of the number of columns found in the file %d" %
                    (len(fields), bestcol))
            names = fields
        else:
            hhhh, _ = 0, bestcol
            while _ > 0:
                hhhh, _ = hhhh, _ / 10  # pylint: disable=W0127
            format = "c%0" + str(hhhh) + "d"
            names = [format % i for i in range(bestcol)]

        for k in columns:
            if k >= len(names):
                raise Exception(
                    "incoherence in the file being read: %d >= %d: " %
                    (k, len(names)) + repr(names) + "\n" + repr(columns))
            columns[k] = (changes.get(names[k], names[k]), columns[k][0])

        self.LOG(
            "  TextFile.guess_columns: header ",
            header,
            " columns ",
            columns)
        coy = columns.copy()

        # end
        exp = self._build_regex(bestsep, columns, regex=regex)
        self.LOG("  TextFile.guess_columns: regex ", exp)

        # determines the length of columns
        length = {}
        no = 0
        for line in lines:
            spl = line.split(bestsep)
            if len(spl) != len(columns):
                continue
            no += 1
            for i, c in enumerate(spl):
                vl = length.get(i, 0)
                if vl < len(c):
                    length[i] = len(c)

        for c in columns:
            v = columns[c]
            if v[1] == str and c in length and length[c] > 0:
                v = (v[0], (v[1], length[c] * 2))
                columns[c] = v

        if coy != columns:
            self.LOG(
                "  TextFile.guess_columns: header ",
                header,
                " columns ",
                columns)

        return header, columns, bestsep, exp

    def count_rejected_lines(self, header, exp, output=None):
        """
        Counts the number of rejected lines by regular expression exp.

        @param      header          header or not in the first line
        @param      exp             regular expression
        @param      output          if != None, output is a stream which will receive the unrecognized line (see below)
        @return                     nb_accepted, nb rejected

        Format for the file containing the unrecognized lines:
        @code
        line number \t  line
        @endcode

        """
        if isinstance(exp, str):
            exp = re.compile(exp, re.U)
        acc, rej = 0., 0.
        temp = TextFile(self.filename, fLOG=self.LOG, encoding=self._encoding)
        temp.open()
        nb = 0
        for line in temp:
            nb += 1
            if header and acc + rej == 0:
                header = False
                continue
            if len(line) == 0:
                continue
            r = exp.search(line)
            if r:
                acc += 1
            else:
                rej += 1
                if output is not None:
                    output.write("%d\t%s\n" % (nb - 1, line))
        temp.close()
        return acc, rej

    _build_regex_default_value_types = {int: "([-]?[1-9][0-9]*?)|(0?)",
                                        decimal.Decimal: "([-]?[1-9][0-9]*?L?)|(0?)",
                                        float: "[-]?[0-9]*?([.][0-9]*?)?([eE][-]?[0-9]{0,4})?",
                                        str: ".*"}

    def _build_regex(self, sep, columns,  # pylint: disable=W0102
                     exp=_build_regex_default_value_types,  # pylint: disable=W0102
                     nomore=False, regex=None):
        """
        Builds a regular expression.

        @param          sep             separator
        @param          columns         columns definition
        @param          exp             regular expression associated to each type, (see below for the default value)
        @param          nomore          private argument, no more try, not possible to simplify
        @param          regex           if the default expression for a field is not the expected one, look into regex if there is one
        @return                         regex

        Default value for ``exp``:

        @code
            {
                int:             "([-]?[1-9][0-9]*?)|(0?)",
                decimal.Decimal: "([-]?[1-9][0-9]*?L?)|(0?)",
                float:           "[-]?[0-9]*?([.][0-9]*?)?([eE][-]?[0-9]{0,4})?",
                str:             ".*"
            }
        @endcode

        """
        if regex is None:
            regex = {}
        mx = max(columns.keys()) + 1
        res = [None for i in range(mx)]
        for k, v in columns.items():
            t = v[1]
            if t not in exp:
                raise Exception("unknown type %s" % str(t))
            nv0 = v[0].strip()
            if nv0 in regex:
                res[k] = (nv0, regex[nv0])
            else:
                res[k] = (nv0, exp[t])
        for c in res:
            if " " in c[0]:
                raise ValueError(
                    "Accents are not allowed for column names: {0}".format(c))
        res = ["(?P<%s>%s)" % c for c in res]
        if sep == "\t":
            sep = "\\t"
        final = "^%s$" % sep.join(res)

        try:
            self.LOG("  compiling", final)
            exp = re.compile(final)
            return final
        except Exception as e:
            if "but this version only supports 100 named groups" in str(e):
                self.LOG(
                    "  problem with expression (more than 100 groups) ",
                    final)
            if nomore:
                if "bad character in group name" in str(e):
                    reg = re.compile("?P<(.*?)>")
                    all = reg.findall(final)
                    s = ",".join(all)
                    raise RuntimeError(  # pylint: disable=W0707
                        "this expression does not compile (%r), pattern %r, columns %r" %
                        (str(e), final, s))
                raise RuntimeError(  # pylint: disable=W0707
                    "This expression does not compile (%r), pattern %r" %
                    (str(e), final))

        exp = {int: "[-]?[0-9]*?",
               float: "[0-9.eE]*?",
               str: ".*"}
        return self._build_regex(sep, columns, exp, True)



"""
@file
@brief Check code style.

.. versionadded:: 1.8
"""
import os
from pyquickhelper.pycode import check_pep8


def _private_test_style_src(fLOG, run_lint, verbose=False, pattern=".*[.]py$"):
    thi = os.path.abspath(os.path.dirname(__file__))
    src_ = os.path.normpath(os.path.join(thi, ".."))
    check_pep8(src_, fLOG=fLOG, neg_pattern='.*((Parser)|(Lexer)|(Listener)|(antlr_grammar_use))[.]py$',
               verbose=verbose, pattern=pattern, run_lint=run_lint,
               pylint_ignore=('C0103', 'C1801', 'R0201', 'R1705', 'W0108', 'W0613',
                              'C0111', 'W0622', 'W0612', 'C0412', 'W0621',
                              'W0212', 'R1704', 'W0622', 'W0201', 'R1710',
                              'W0703', 'R1703', 'R0911', 'R0912', 'R0915',
                              'E0203', 'C0302', 'C0200', 'R1702', 'E1101',
                              'R0914', 'W0123', 'W0123', 'W0107', 'C0415'),
               skip=["http_retrieve.py:191: W0703",
                     "astock.py:135: W0703",
                     "astock.py:229: W0703",
                     "http_retrieve.py:192: W0703",
                     "Class 'DatabaseCore' has no 'SCRIPT_LOOKUP'",
                     "sql_interface.py:18: W0231",
                     "file_text_binary.py:680: W0102",
                     "sql_interface_database.py:20: W0231",
                     "table_formula.py:108: W0631",
                     "R0401: Cyclic import (src.pyensae.sql.sql_interface",
                     "R0401: Cyclic import (pyensae.sql.sql_interface",
                     "database_core2.py:230: R1714",
                     "RFilter.py",
                     ])


def _private_test_style_test(fLOG, run_lint, verbose=False, pattern=".*[.]py$"):
    thi = os.path.abspath(os.path.dirname(__file__))
    test_ = os.path.normpath(os.path.join(thi, "..", "..", '_unittests'))
    check_pep8(test_, fLOG=fLOG, neg_pattern="(temp_.*)|(.*test_parse_code.*[.]py)",
               verbose=verbose, pattern=pattern, run_lint=run_lint,
               pylint_ignore=('C0103', 'C1801', 'R0201', 'R1705', 'W0108', 'W0613',
                              'C0111', 'W0622', 'W0612', 'C0412', 'W0621',
                              'W0125', 'E1127', 'E1101', 'W1402', 'W0212',
                              'C0415'),
               skip=[])
